#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
#
##############################################################################

"""
Parsing of NEWICK-format tree from a stream.
"""

import re
import warnings
from dendropy.utility import error
from dendropy.utility import deprecate
from dendropy.utility.textprocessing import StringIO
from dendropy.dataio import tokenizer
from dendropy.dataio import nexusprocessing
from dendropy.dataio import ioservice

##############################################################################
## NewickReader

class NewickReader(ioservice.DataReader):
    """
    Parser for NEWICK-formatted data.
    """

    _default_rooting_directive = None
    _default_tree_weight = 1.0

    class NewickReaderError(error.DataParseError):
        def __init__(self, message,
                line_num=None,
                col_num=None,
                stream=None):
            error.DataParseError.__init__(self,
                    message=message,
                    line_num=line_num,
                    col_num=col_num,
                    stream=stream)

    class NewickReaderInvalidTokenError(NewickReaderError):
        def __init__(self,
                message,
                line_num=None,
                col_num=None,
                stream=None):
            NewickReader.NewickReaderError.__init__(self,
                    message=message,
                    line_num=line_num,
                    col_num=col_num,
                    stream=stream)

    class NewickReaderMalformedStatementError(NewickReaderError):
        def __init__(self,
                message,
                line_num=None,
                col_num=None,
                stream=None):
            NewickReader.NewickReaderError.__init__(self,
                    message=message,
                    line_num=line_num,
                    col_num=col_num,
                    stream=stream)

    class NewickReaderIncompleteTreeStatementError(NewickReaderMalformedStatementError):
        def __init__(self,
                message,
                line_num=None,
                col_num=None,
                stream=None):
            NewickReader.NewickReaderMalformedStatementError.__init__(self,
                    message=message,
                    line_num=line_num,
                    col_num=col_num,
                    stream=stream)

    class NewickReaderInvalidValueError(NewickReaderError):
        def __init__(self,
                message,
                line_num=None,
                col_num=None,
                stream=None):
            NewickReader.NewickReaderError.__init__(self,
                    message=message,
                    line_num=line_num,
                    col_num=col_num,
                    stream=stream)

    class NewickReaderDuplicateTaxonError(NewickReaderError):
        def __init__(self,
                message,
                line_num=None,
                col_num=None,
                stream=None):
            detailed = ("Multiple occurrences of the same taxa on trees are not"
            " supported: trees with duplicate node labels can only be"
            " processed if the labels are not parsed as operational taxonomic"
            " unit concepts but instead as simply node labels by specifying"
            " 'suppress_internal_node_taxa=True, suppress_leaf_node_taxa=True'."
            " Duplicate taxon labels: {}").format(message)
            NewickReader.NewickReaderError.__init__(self,
                    message=detailed,
                    line_num=line_num,
                    col_num=col_num,
                    stream=stream)

    def __init__(self, **kwargs):
        """Keyword Arguments
        -----------------
        rooting : string, {['default-unrooted'], 'default-rooted', 'force-unrooted', 'force-rooted'}
            Specifies how trees in the data source should be intepreted with
            respect to their rooting:

                'default-unrooted' [default]:
                    All trees are interpreted as unrooted unless a '[&R]'
                    comment token explicitly specifies them as rooted.
                'default-rooted'
                    All trees are interpreted as rooted unless a '[&U]'
                    comment token explicitly specifies them as unrooted.
                'force-unrooted'
                    All trees are unconditionally interpreted as unrooted.
                'force-rooted'
                    All trees are unconditionally interpreted as rooted.

        edge_length_type : type, default: ``float``
            Specifies the type of the edge lengths (``int`` or ``float``). Tokens
            interpreted as branch lengths will be cast to this type.
            Defaults to ``float``.
        suppress_edge_lengths : boolean, default: |False|
            If |True|, edge length values will not be processed. If |False|,
            edge length values will be processed.
        extract_comment_metadata : boolean, default: |True|
            If |True| (default), any comments that begin with '&' or '&&' will
            be parsed and stored as part of the annotation set of the
            corresponding object (accessible through the ``annotations``
            attribute of the object). This requires that the comment
            contents conform to a particular format (NHX or BEAST: 'field =
            value'). If |False|, then the comments will not be parsed,
            but will be instead stored directly as elements of the ``comments``
            list attribute of the associated object.
        store_tree_weights : boolean, default: |False|
            If |True|, process the tree weight (e.g. "[&W 1/2]") comment
            associated with each tree, if any. Defaults to |False|.
        finish_node_fn : function object, default: |None|
            If specified, this function will be applied to each node after
            it has been constructed.
        case_sensitive_taxon_labels : boolean, default: |False|
            If |True|, then taxon labels are case sensitive (e.g., "P.regius"
            and "P.REGIUS" wil be treated as different operation taxonomic
            unit concepts). Otherwise, taxon label intepretation will be made
            without regard for case.
        preserve_underscores : boolean, default: |False|
            If |True|, unquoted underscores in labels will *not* converted to
            spaces. Defaults to |False|: all underscores not protected by
            quotes will be converted to spaces.
        suppress_internal_node_taxa : boolean, default: |True|
            If |False|, internal node labels will be instantantiated into
            |Taxon| objects. If |True|, internal node labels
            will *not* be instantantiated as strings.
        suppress_leaf_node_taxa : boolean, default: |False|
            If |False|, leaf (external) node labels will be instantantiated
            into |Taxon| objects. If |True|, leaff (external) node
            labels will *not* be instantantiated as strings.
        is_parse_jplace_tokens : boolean: |False|
            If |True|, then accept edge numbering according to the jplace
            format, as described in Matsen et. al. PLoS One, 2012
            http://dx.doi.org/10.1371/journal.pone.0031009. An instance variable
            edge_index is added to the returned tree, and an edge_number is
            added to each edge. If False [default], encountering edge labels
            raises a NewickReaderMalformedStatementError.
        is_assign_internal_labels_to_edges : boolean, default: |None|
            If |True|, internal node labels will be assigned as edge labels.
        terminating_semicolon_required : boolean, default: |True|
            If |True| [default], then a tree statement that does not end in a
            semi-colon is an error. If |False|, then no error will be raised.
        ignore_unrecognized_keyword_arguments : boolean, default: |False|
            If |True|, then unsupported or unrecognized keyword arguments will
            not result in an error. Default is |False|: unsupported keyword
            arguments will result in an error.

        """

        # base
        ioservice.DataReader.__init__(self)

        self._rooting = None
        ## (TEMPORARY and UGLY!!!!) Special handling for legacy signature
        if "as_unrooted" in kwargs or "as_rooted" in kwargs or "default_as_rooted" in kwargs or "default_as_unrooted" in kwargs:
            import collections
            legacy_kw = ("as_unrooted", "as_rooted", "default_as_rooted", "default_as_unrooted")
            legacy_kw_str = ", ".join("'{}'".format(k) for k in legacy_kw)
            if "rooting" in kwargs:
                raise ValueError("Cannot specify 'rooting' keyword argument in conjunction with any of the (legacy) keyword arguments ({}). Use 'rooting' alone.".format(legacy_kw_str))
            specs = collections.Counter(k for k in kwargs.keys() if k in legacy_kw)
            if sum(specs.values()) > 1:
                raise ValueError("Cannot specify more than one of {{ {} }} at the same time".format(legacy_kw_str))
            kw = list(specs.keys())[0]
            if kw == "as_unrooted":
                if kwargs[kw]:
                    corrected = "force-unrooted"
                else:
                    corrected = "force-rooted"
            elif kw == "as_rooted":
                if kwargs[kw]:
                    corrected = "force-rooted"
                else:
                    corrected = "force-unrooted"
            elif kw == "default_as_unrooted":
                if kwargs[kw]:
                    corrected = "default-unrooted"
                else:
                    corrected = "default-rooted"
            elif kw == "default_as_rooted":
                if kwargs[kw]:
                    corrected = "default-rooted"
                else:
                    corrected = "default-unrooted"
            msg = StringIO()
            # error.critical_deprecation_alert("\n{}\nUse of keyword argument '{}={}' is deprecated; use 'rooting=\"{}\"' instead".format(msg.getvalue(), kw, kwargs[kw], corrected),
            #         stacklevel=4)
            deprecate.dendropy_deprecation_warning(
                    preamble="Deprecated since DendroPy 4:",
                    old_construct="{}={}".format(kw, kwargs[kw]),
                    new_construct="rooting='{}'".format(corrected),
                    stacklevel=7)
            kwargs.pop(kw)
            kwargs["rooting"] = corrected
        if "allow_duplicate_taxon_labels" in kwargs:
            raise ValueError(
                "'allow_duplicate_taxon_labels' is no longer"
                " supported: trees with duplicate node labels can only be"
                " processed if the labels are not parsed as operational taxonomic"
                " unit concepts but instead as simply node labels by specifying"
                " 'suppress_internal_node_taxa=True, suppress_leaf_node_taxa=True'."
            )
        # self.rooting = kwargs.pop("rooting", "default-unrooted")
        self.rooting = kwargs.pop("rooting", self.__class__._default_rooting_directive)
        self.edge_length_type = kwargs.pop("edge_length_type", float)
        self.suppress_edge_lengths = kwargs.pop("suppress_edge_lengths", False)
        self.extract_comment_metadata = kwargs.pop('extract_comment_metadata', True)
        self.store_tree_weights = kwargs.pop("store_tree_weights", False)
        self.default_tree_weight = kwargs.pop("default_tree_weight", self.__class__._default_tree_weight)
        self.finish_node_fn = kwargs.pop("finish_node_fn", None)
        self.case_sensitive_taxon_labels = kwargs.pop('case_sensitive_taxon_labels', False)
        self.preserve_unquoted_underscores = kwargs.pop('preserve_underscores', False)
        self.suppress_internal_node_taxa = kwargs.pop("suppress_internal_node_taxa", True)
        self.suppress_leaf_node_taxa = kwargs.pop("suppress_external_node_taxa", False) # legacy (will be deprecated)
        self.suppress_leaf_node_taxa = kwargs.pop("suppress_leaf_node_taxa", self.suppress_leaf_node_taxa)
        self.is_parse_jplace_tokens = kwargs.pop("is_parse_jplace_tokens", False)
        self.is_assign_internal_labels_to_edges = kwargs.pop("is_assign_internal_labels_to_edges", None)
        if self.is_assign_internal_labels_to_edges and not self.suppress_internal_node_taxa:
            raise ValueError("Conflicting options: cannot simultaneously assign internal labels to edges and to internal taxa")
        self.terminating_semicolon_required = kwargs.pop("terminating_semicolon_required", True)
        self.check_for_unused_keyword_arguments(kwargs)

        # per-tree book-keeping
        self._tree_statement_complete = None
        self._parenthesis_nesting_level = None
        self._seen_taxa = None

    def tree_iter(self,
            stream,
            taxon_symbol_mapper,
            tree_factory):
        """
        Iterator that yields trees in NEWICK-formatted source.

        Parameters
        ----------
        stream : file or file-like object
            A file or file-like object opened for reading.
        taxon_namespace : |TaxonNamespace|
            Operational taxonomic unit namespace to use for taxon management.
        tree_factory : function object
            A function that returns a new |Tree| object when called
            without arguments.

        Returns
        -------
        iter : :py`collections.Iterator` [|Tree|]
            An iterator yielding |Tree| objects constructed based on
            data in ``stream``.
        """
        nexus_tokenizer = nexusprocessing.NexusTokenizer(stream,
                preserve_unquoted_underscores=self.preserve_unquoted_underscores)
        while True:
            tree = self._parse_tree_statement(
                    nexus_tokenizer=nexus_tokenizer,
                    tree_factory=tree_factory,
                    taxon_symbol_map_fn=taxon_symbol_mapper.require_taxon_for_symbol)
            yield tree
            if tree is None:
                # raise StopIteration
                return

    def _read(self,
            stream,
            taxon_namespace_factory=None,
            tree_list_factory=None,
            char_matrix_factory=None,
            state_alphabet_factory=None,
            global_annotations_target=None):
        taxon_namespace = taxon_namespace_factory(label=None)
        tree_list = tree_list_factory(label=None, taxon_namespace=taxon_namespace)
        taxon_symbol_mapper = nexusprocessing.NexusTaxonSymbolMapper(
                taxon_namespace=taxon_namespace,
                enable_lookup_by_taxon_number=False,
                case_sensitive=self.case_sensitive_taxon_labels)
        tree_factory = tree_list.new_tree
        for tree in self.tree_iter(stream=stream,
                taxon_symbol_mapper=taxon_symbol_mapper,
                tree_factory=tree_factory):
            pass
        product = self.Product(
                taxon_namespaces=None,
                tree_lists=[tree_list],
                char_matrices=None)
        return product

    def _get_rooting(self):
        """
        Get rooting interpretation configuration.
        """
        return self._rooting
    def _set_rooting(self, val):
        """
        Set rooting interpretation configuration.
        """
        if val not in ("force-unrooted", "force-rooted", "default-unrooted", "default-rooted", None,):
            raise ValueError("Unrecognized rooting directive: '{}'".format(val))
        self._rooting = val
    rooting = property(_get_rooting, _set_rooting)

    def _parse_tree_statement(self,
            nexus_tokenizer,
            tree_factory,
            taxon_symbol_map_fn):
        """
        Parses a single tree statement from a token stream and constructs a
        corresponding Tree object. Expects that the first non-comment and
        non-semi-colon token to be found, including the current token, to be
        the parenthesis that opens the tree statement. When complete, the
        current token will be the token immediately following the semi-colon,
        if any.
        """
        current_token = nexus_tokenizer.current_token
        tree_comments = nexus_tokenizer.pull_captured_comments()
        while (current_token == ";" or current_token is None) and not nexus_tokenizer.is_eof():
            current_token = nexus_tokenizer.require_next_token()
            tree_comments = nexus_tokenizer.pull_captured_comments()
        if nexus_tokenizer.is_eof():
            return None
        if current_token != "(":
            # allow for possibility of single node tree, e.g.: T0:10;
            self._parenthesis_nesting_level = 0
            # raise NewickReader.NewickReaderInvalidTokenError(
            #         message="Expecting '{}' but found '{}'".format("(", current_token),
            #         line_num=nexus_tokenizer.token_line_num,
            #         col_num=nexus_tokenizer.token_column_num,
            #         stream=nexus_tokenizer.src)
        else:
            self._parenthesis_nesting_level = 1
        tree = tree_factory()
        self._process_tree_comments(tree, tree_comments, nexus_tokenizer)
        self._tree_statement_complete = False
        self._seen_taxa = set()
        self._parse_tree_node_description(
                nexus_tokenizer=nexus_tokenizer,
                tree=tree,
                current_node=tree.seed_node,
                taxon_symbol_map_fn=taxon_symbol_map_fn,
                is_internal_node=None)
        current_token = nexus_tokenizer.current_token
        if not self._tree_statement_complete:
            raise NewickReader.NewickReaderIncompleteTreeStatementError(
                    message="Incomplete or improperly-terminated tree statement (last character read was '{}' instead of a semi-colon ';')".format(nexus_tokenizer.current_token),
                    line_num=nexus_tokenizer.token_line_num,
                    col_num=nexus_tokenizer.token_column_num,
                    stream=nexus_tokenizer.src)
        self._seen_taxa = None
        self._parenthesis_nesting_level = None
        self._tree_statement_complete = None
        while current_token == ";" and not nexus_tokenizer.is_eof():
            nexus_tokenizer.clear_captured_comments()
            current_token = nexus_tokenizer.next_token()
        return tree

    def _process_tree_comments(self, tree, tree_comments, nexus_tokenizer):
        # NOTE: this also unconditionally sets the tree rootedness and
        # weighting if no comment indicating these are found; for this to work
        # in the current implementation, this method must be called once and
        # exactly once per tree.
        if not tree_comments:
            tree.is_rooted = self._parse_tree_rooting_state("")
            if self.store_tree_weights:
                tree.weight = self.default_tree_weight
            return
        rooting_token_found = False
        weighting_token_found = False
        for comment in tree_comments:
            stripped_comment = comment.strip()
            if stripped_comment in ["&u", "&U", "&r", "&R"]:
                tree.is_rooted = self._parse_tree_rooting_state(stripped_comment)
                rooting_token_found = True
            elif (self.store_tree_weights
                    and (stripped_comment.startswith("&W ") or stripped_comment.startswith("&w "))
                    ):
                try:
                    weight_expression = stripped_comment[2:]
                    if not weight_expression:
                        raise ValueError
                    we_parts = weight_expression.split("/")
                    if len(we_parts) > 2:
                        raise ValueError
                        # raise NewickReader.NewickReaderInvalidValueError(
                        #         message="Invalid tree weight expression: '{}'".format(weight_expression),
                        #         line_num=nexus_tokenizer.token_line_num,
                        #         col_num=nexus_tokenizer.token_column_num,
                        #         stream=nexus_tokenizer.src)
                    elif len(we_parts) == 2:
                        x = float(we_parts[0])
                        y = float(we_parts[1])
                        tree.weight = x/y
                    else:
                        tree.weight = float(we_parts[0])
                    weighting_token_found = True
                except ValueError:
                    exc = NewickReader.NewickReaderInvalidValueError(
                            message="Invalid tree weight expression: '{}'".format(stripped_comment),
                            line_num=nexus_tokenizer.token_line_num,
                            col_num=nexus_tokenizer.token_column_num,
                            stream=nexus_tokenizer.src)
                    exc.__context__ = None # Python 3.0, 3.1, 3.2
                    exc.__cause__ = None # Python 3.3, 3.4
                    raise exc
            elif self.extract_comment_metadata and comment.startswith("&"):
                annotations = nexusprocessing.parse_comment_metadata_to_annotations(
                    comment=comment)
                if annotations:
                    tree.annotations.update(annotations)
                else:
                    tree.comments.append(comment)
            else:
                tree.comments.append(comment)
        if not rooting_token_found:
            tree.is_rooted = self._parse_tree_rooting_state("")
        if self.store_tree_weights and not weighting_token_found:
            tree.weight = self.default_tree_weight

    def _parse_tree_rooting_state(self, rooting_comment=None):
        """
        Returns rooting state for tree with given rooting comment token, taking
        into account ``rooting`` configuration.
        """
        if self._rooting == "force-unrooted":
            return False
        elif self._rooting == "force-rooted":
            return True
        elif rooting_comment == "&R" or rooting_comment == "&r":
            return True
        elif rooting_comment == "&U" or rooting_comment == "&u":
            return False
        elif self._rooting == "default-rooted":
            return True
        elif self._rooting == "default-unrooted":
            return False
        elif self._rooting is None:
            return None
        else:
            raise TypeError("Unrecognized rooting directive: '{}'".format(self._rooting))

    def _parse_tree_node_description(
            self,
            nexus_tokenizer,
            tree,
            current_node,
            taxon_symbol_map_fn,
            is_internal_node=None):
        """
        Assuming that the iterator is currently sitting on a parenthesis that
        opens a node with children or the label of a leaf node, this will
        populate the node ``node`` appropriately (label, edge length, comments,
        metadata etc.) and recursively parse and add the node's
        children. When complete, the token will be the token immediately
        following the end of the node or tree statement if this is the root
        node, i.e. the token *following* the closing parenthesis of the node or
        the semi-colon terminating a tree statement.
        """
        current_node_comments = nexus_tokenizer.pull_captured_comments()
        if nexus_tokenizer.current_token == "(":
            # self._parenthesis_nesting_level += 1 # handled by calling code
            nexus_tokenizer.require_next_token()
            node_created = False
            while True:
                if nexus_tokenizer.current_token == ",":
                    if not node_created: #184
                        # no node has been created yet: ',' designates a
                        # preceding blank node
                        new_node = tree.node_factory()
                        nexusprocessing.process_comments_for_item(item=new_node,
                                item_comments=nexus_tokenizer.pull_captured_comments(),
                                extract_comment_metadata=self.extract_comment_metadata)
                        self._finish_node(new_node)
                        current_node.add_child(new_node)
                        ## node_created = True # do not flag node as created to allow for an extra node to be created in the event of (..,)
                    nexus_tokenizer.require_next_token()
                    while nexus_tokenizer.current_token == ",": #192
                        # another blank node
                        new_node = tree.node_factory()
                        nexusprocessing.process_comments_for_item(item=new_node,
                                item_comments=nexus_tokenizer.pull_captured_comments(),
                                extract_comment_metadata=self.extract_comment_metadata)
                        self._finish_node(new_node)
                        current_node.add_child(new_node)
                        # node_created = True; # do not flag node as created: extra node needed in the event of (..,)
                        nexus_tokenizer.require_next_token()
                    if not node_created and nexus_tokenizer.current_token == ")": #200
                        # end of node
                        new_node = tree.node_factory();
                        nexusprocessing.process_comments_for_item(item=new_node,
                                item_comments=nexus_tokenizer.pull_captured_comments(),
                                extract_comment_metadata=self.extract_comment_metadata)
                        self._finish_node(new_node)
                        current_node.add_child(new_node)
                        node_created = True;
                elif nexus_tokenizer.current_token == ")": #206
                    # end of child nodes
                    self._parenthesis_nesting_level -= 1
                    nexus_tokenizer.require_next_token()
                    break
                else: #210
                    # assume child nodes: a leaf node (if a label) or
                    # internal (if a parenthesis)
                    if nexus_tokenizer.current_token == "(":
                        self._parenthesis_nesting_level += 1
                        is_new_internal_node = True
                    else:
                        is_new_internal_node = False
                    new_node = tree.node_factory();
                    nexusprocessing.process_comments_for_item(item=new_node,
                            item_comments=nexus_tokenizer.pull_captured_comments(),
                            extract_comment_metadata=self.extract_comment_metadata)
                    self._parse_tree_node_description(
                            nexus_tokenizer=nexus_tokenizer,
                            tree=tree,
                            current_node=new_node,
                            taxon_symbol_map_fn=taxon_symbol_map_fn,
                            is_internal_node=is_new_internal_node,
                            )
                    current_node.add_child(new_node);
                    node_created = True;
        label_parsed = False
        self._tree_statement_complete = False
        if is_internal_node is None:
            # Initial call using ``seed_node`` does not set ``is_internal_node`` to
            # |True| or |False|, explicitly, but rather |None|. If this is the
            # case, the rest of the tree has be constructed, and we simply look
            # at whether there are children or not to determine if it is an
            # internal node. This approach allows for a single-tip tree.
            if current_node._child_nodes:
                is_internal_node = True
        if current_node_comments is None:
            current_node_comments = []
        while True:
            cc = nexus_tokenizer.pull_captured_comments()
            if cc is not None:
                current_node_comments.extend(cc)
            if nexus_tokenizer.current_token == ":": #246
                nexus_tokenizer.require_next_token()
                if not self.suppress_edge_lengths:
                    try:
                        edge_length = self.edge_length_type(nexus_tokenizer.current_token)
                    except ValueError:
                        raise NewickReader.NewickReaderMalformedStatementError(
                                message="Invalid edge length: '{}'".format(nexus_tokenizer.current_token),
                                line_num=nexus_tokenizer.token_line_num,
                                col_num=nexus_tokenizer.token_column_num,
                                stream=nexus_tokenizer.src)
                    current_node.edge.length = edge_length
                try:
                    nexus_tokenizer.require_next_token()
                except tokenizer.Tokenizer.UnexpectedEndOfStreamError as e:
                    if self.terminating_semicolon_required:
                        message = e.message + ". (Perhaps the terminating semicolon for the tree statement is missing? If so, add a semicolon to the tree statement or specify 'terminating_semicolon_required=False' to allow for missing semicolons)"
                        raise tokenizer.Tokenizer.UnexpectedEndOfStreamError(
                                message=message,
                                line_num=e.line_num,
                                col_num=e.col_num,
                                stream=e.stream)
                    else:
                        self._tree_statement_complete = True
                        break

            elif nexus_tokenizer.current_token == ")": #253
                # closing of parent token
                # self._parenthesis_nesting_level -= 1 # handled by calling code
                nexusprocessing.process_comments_for_item(item=current_node,
                        item_comments=current_node_comments,
                        extract_comment_metadata=self.extract_comment_metadata)
                self._finish_node(current_node)
                return current_node
            elif nexus_tokenizer.current_token == ";": #256
                # end of tree statement
                self._tree_statement_complete = True
                nexus_tokenizer.next_token()
                break
            elif nexus_tokenizer.current_token == ",": #260
                # end of this node
                nexusprocessing.process_comments_for_item(item=current_node,
                            item_comments=current_node_comments,
                            extract_comment_metadata=self.extract_comment_metadata)
                self._finish_node(current_node)
                return current_node
            elif nexus_tokenizer.current_token == "(": #263
                # start of another node or tree without finishing this
                # node
                self._parenthesis_nesting_level += 1
                raise NewickReader.NewickReaderMalformedStatementError(
                        message="Malformed tree statement",
                        line_num=nexus_tokenizer.token_line_num,
                        col_num=nexus_tokenizer.token_column_num,
                        stream=nexus_tokenizer.src)
            elif self.is_parse_jplace_tokens and nexus_tokenizer.current_token == '{':
                # Edge number from .jplace format
                nexus_tokenizer.require_next_token()
                edge_number = int(nexus_tokenizer.current_token)
                edge = current_node.edge
                edge.edge_number = edge_number
                try:
                    tree.edge_index.insert(edge_number, edge)
                except AttributeError:
                    tree.edge_index = []
                    tree.edge_index.insert(edge_number, edge)
                nexus_tokenizer.require_next_token() # for closing '}'
                nexus_tokenizer.require_next_token()
            else: #267
                # label
                if label_parsed: #269
                    msg = "Expecting ':'"
                    if self.is_parse_jplace_tokens:
                        msg += ", '{'"
                    msg += ", ')', ',' or ';' after reading label but found '{}'".format(nexus_tokenizer.current_token)
                    raise NewickReader.NewickReaderMalformedStatementError(
                            message=msg,
                            line_num=nexus_tokenizer.token_line_num,
                            col_num=nexus_tokenizer.token_column_num,
                            stream=nexus_tokenizer.src)
                else:
                    # Label
                    label = nexus_tokenizer.current_token
                    if ( (is_internal_node and self.suppress_internal_node_taxa)
                            or ((not is_internal_node) and self.suppress_leaf_node_taxa) ):
                        if self.is_assign_internal_labels_to_edges:
                            current_node.edge.label = label
                        else:
                            current_node.label = label
                    else:
                        node_taxon = taxon_symbol_map_fn(label)
                        if node_taxon in self._seen_taxa:
                            raise NewickReader.NewickReaderDuplicateTaxonError(
                                    message=node_taxon.label,
                                    line_num=nexus_tokenizer.token_line_num,
                                    col_num=nexus_tokenizer.token_column_num,
                                    stream=nexus_tokenizer.src)
                        self._seen_taxa.add(node_taxon)
                        current_node.taxon = node_taxon
                    label_parsed = True;
                    # nexus_tokenizer.require_next_token()
                    try:
                        nexus_tokenizer.require_next_token()
                    except tokenizer.Tokenizer.UnexpectedEndOfStreamError:
                        if self.terminating_semicolon_required:
                            raise
                        else:
                            break
        ## if we are here, we have reached the end of the tree
        if self._parenthesis_nesting_level != 0:
            raise NewickReader.NewickReaderMalformedStatementError(
                    message="Unbalanced parentheses at tree statement termination: balance index = {}".format(self._parenthesis_nesting_level),
                    line_num=nexus_tokenizer.token_line_num,
                    col_num=nexus_tokenizer.token_column_num,
                    stream=nexus_tokenizer.src)
        nexusprocessing.process_comments_for_item(item=current_node,
                item_comments=current_node_comments,
                extract_comment_metadata=self.extract_comment_metadata)
        self._finish_node(current_node)
        return current_node

    def _finish_node(self, node):
        if self.finish_node_fn is not None:
            self.finish_node_fn(node)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Models, modeling and model-fitting of the protracted speciation, as described in::

        Etienne, R.S., Morlon, H., and Lambert, A. 2014. Estimating the
        duration of speciation from phylogenies. Evolution 2014: 2430-2440.
        doi:10.1111/evo.12433

"""

import math
import heapq
import dendropy
import itertools
from dendropy.utility import GLOBAL_RNG
from dendropy.utility.error import ProcessFailedException
from dendropy.utility.error import TreeSimTotalExtinctionException
from dendropy.calculate import probability

def _D(speciation_initiation_rate,
       speciation_completion_rate,
       incipient_species_extinction_rate):
    r"""
    Returns value of D, as given in eq. 5 in Etienne et al.
    (2014).

    Parameters
    ----------

    speciation_initiation_rate : float
        The birth rate, b (the incipient species birth
        rate and the "good" species birth rate are assumed to be equal):
        the rate at which new (incipient) species are produced from
        either incipient or "good" species lineages.
    speciation_completion_rate : float
        The rate at which incipient species get converted to good
        species, $lambda_1$.
    incipient_species_extinction_rate : float
        The incipient species exctinction rate, $\mu_1$: the rate at which
        incipient species go extinct.

    Returns
    -------
    t : float
        The duration of speciation.

    """
    D = math.sqrt(
            pow(speciation_completion_rate + speciation_initiation_rate - incipient_species_extinction_rate, 2)
            + (4.0 * speciation_completion_rate * incipient_species_extinction_rate)
        )
    return D

def _phi(speciation_initiation_rate,
       speciation_completion_rate,
       incipient_species_extinction_rate):
    r"""
    Returns value of $\varphi$, as given in eq. 6 in Etienne et al.
    (2014).

    Parameters
    ----------

    speciation_initiation_rate : float
        The birth rate, b (the incipient species birth
        rate and the "good" species birth rate are assumed to be equal):
        the rate at which new (incipient) species are produced from
        either incipient or "good" species lineages.
    speciation_completion_rate : float
        The rate at which incipient species get converted to good
        species, $lambda_1$.
    incipient_species_extinction_rate : float
        The incipient species exctinction rate, $\mu_1$: the rate at which
        incipient species go extinct.

    Returns
    -------
    t : float
        The duration of speciation.

    """
    phi = speciation_completion_rate - speciation_initiation_rate + incipient_species_extinction_rate
    return phi

def expected_duration_of_speciation(
        speciation_initiation_rate,
        speciation_completion_rate,
        incipient_species_extinction_rate,
        D=None,
        ):
    r"""
    Returns mean duration of speciation, following Eqs. 4 in Etienne et al.
    (2014):

        The duration of speciation differs from the speciation-completion
        time in that the latter is the waiting time until a single
        incipient lineage completes the speciation process if extinction
        was zero, whereas the former is the time needed for an incipient
        species or one of its descendants to complete speciation, condi-
        tional on the fact that speciation completes, that is, this is the
        time taken by any species that succeeded in speciating completely.

    Parameters
    ----------

    speciation_initiation_rate : float
        The birth rate, b (the incipient species birth
        rate and the "good" species birth rate are assumed to be equal):
        the rate at which new (incipient) species are produced from
        either incipient or "good" species lineages.
    speciation_completion_rate : float
        The rate at which incipient species get converted to good
        species, $lambda_1$.
    incipient_species_extinction_rate : float
        The incipient species exctinction rate, $\mu_1$: the rate at which
        incipient species go extinct.
    D : float
        Value of ``D`` (as given in Eq. 5 in Etienne et al. 2014). Will be
        calculated if not specified.

    Returns
    -------
    t : float
        The duration of speciation.

    """
    if D is None:
        D = _D(
            speciation_initiation_rate=speciation_initiation_rate,
            speciation_completion_rate=speciation_completion_rate,
            incipient_species_extinction_rate=incipient_species_extinction_rate)
    t1 = 2.0/(D - speciation_completion_rate + speciation_initiation_rate - incipient_species_extinction_rate )
    t2 = math.log(2.0/(1+((speciation_completion_rate - speciation_initiation_rate + incipient_species_extinction_rate)/D)))
    t = t1 * t2
    return t

def probability_of_duration_of_speciation(
        tau,
        speciation_initiation_rate,
        speciation_completion_rate,
        incipient_species_extinction_rate,
        D=None,
        phi=None,
        ):
    r"""
    Returns probability of duration of speciation, tau, following Eqs. 6
    in Etienne et al.

    Parameters
    ----------

    tau : float
        The duration of speciation.
    speciation_initiation_rate : float
        The birth rate, b (the incipient species birth
        rate and the "good" species birth rate are assumed to be equal):
        the rate at which new (incipient) species are produced from
        either incipient or "good" species lineages.
    speciation_completion_rate : float
        The rate at which incipient species get converted to good
        species, $lambda_1$.
    incipient_species_extinction_rate : float
        The incipient species exctinction rate, $\mu_1$: the rate at which
        incipient species go extinct.
    D : float
        Value of ``D`` (as given in Eq. 5 in Etienne et al. 2014). Will be
        calculated if not specified.
    phi : float
        Value of ``phi`` (as given in Eq. 7 in Etienne et al. 2014). Will be
        calculated if not specified.

    Returns
    -------
    p : float
        The probability of the duration of speciation, tau.

    """
    if D is None:
        D = _D(
            speciation_initiation_rate=speciation_initiation_rate,
            speciation_completion_rate=speciation_completion_rate,
            incipient_species_extinction_rate=incipient_species_extinction_rate)
    if phi is None:
        phi = _phi(
            speciation_initiation_rate=speciation_initiation_rate,
            speciation_completion_rate=speciation_completion_rate,
            incipient_species_extinction_rate=incipient_species_extinction_rate)
    n1 = 2.0 * pow(D, 2) * math.exp(-D * tau) * (D + phi)
    d1 = pow(D + phi + math.exp(-D * tau) * (D-phi), 2)
    return n1/d1

def log_probability_of_duration_of_speciation(
        tau,
        speciation_initiation_rate,
        speciation_completion_rate,
        incipient_species_extinction_rate,
        D=None,
        phi=None,
        ):
    r"""
    Returns probability of duration of speciation, tau, following Eqs. 6
    in Etienne et al.

    Parameters
    ----------

    tau : float
        The duration of speciation.
    speciation_initiation_rate : float
        The birth rate, b (the incipient species birth
        rate and the "good" species birth rate are assumed to be equal):
        the rate at which new (incipient) species are produced from
        either incipient or "good" species lineages.
    speciation_completion_rate : float
        The rate at which incipient species get converted to good
        species, $lambda_1$.
    incipient_species_extinction_rate : float
        The incipient species exctinction rate, $\mu_1$: the rate at which
        incipient species go extinct.
    D : float
        Value of ``D`` (as given in Eq. 5 in Etienne et al. 2014). Will be
        calculated if not specified.
    phi : float
        Value of ``phi`` (as given in Eq. 7 in Etienne et al. 2014). Will be
        calculated if not specified.

    Returns
    -------
    p : float
        The probability of the duration of speciation, tau.

    """
    if D is None:
        D = _D(
            speciation_initiation_rate=speciation_initiation_rate,
            speciation_completion_rate=speciation_completion_rate,
            incipient_species_extinction_rate=incipient_species_extinction_rate)
    if phi is None:
        phi = _phi(
            speciation_initiation_rate=speciation_initiation_rate,
            speciation_completion_rate=speciation_completion_rate,
            incipient_species_extinction_rate=incipient_species_extinction_rate)
    n1 = math.log(2.0) + (2 * math.log(D)) - (D * tau) + math.log(D + phi)
    d1 = 2 * (math.log(D + phi + math.exp(-D * tau)*(D-phi)))
    return n1 - d1

def maximum_probability_duration_of_speciation(
        speciation_initiation_rate,
        speciation_completion_rate,
        incipient_species_extinction_rate,
        D=None,
        phi=None,
        ):
    r"""
    Returns duration of speciation that maximizes probability under given
    process parameters, following eq. 8 of Etienne et al (2014).

    Parameters
    ----------

    speciation_initiation_rate : float
        The birth rate, b (the incipient species birth
        rate and the "good" species birth rate are assumed to be equal):
        the rate at which new (incipient) species are produced from
        either incipient or "good" species lineages.
    speciation_completion_rate : float
        The rate at which incipient species get converted to good
        species, $lambda_1$.
    incipient_species_extinction_rate : float
        The incipient species exctinction rate, $\mu_1$: the rate at which
        incipient species go extinct.
    D : float
        Value of ``D`` (as given in Eq. 5 in Etienne et al. 2014). Will be
        calculated if not specified.
    phi : float
        Value of ``phi`` (as given in Eq. 7 in Etienne et al. 2014). Will be
        calculated if not specified.

    Returns
    -------
    t : float
        The duration of speciation with the maximum probability under the
        given process parameters.

    """
    if D is None:
        D = _D(
            speciation_initiation_rate=speciation_initiation_rate,
            speciation_completion_rate=speciation_completion_rate,
            incipient_species_extinction_rate=incipient_species_extinction_rate)
    if phi is None:
        phi = _phi(
            speciation_initiation_rate=speciation_initiation_rate,
            speciation_completion_rate=speciation_completion_rate,
            incipient_species_extinction_rate=incipient_species_extinction_rate)
    x = 1.0/D * math.log((D-phi)/(D+phi))
    return max(0, x)

class ProtractedSpeciationProcess(object):

    class _Lineage(object):

        @classmethod
        def from_pbd_entry(cls, pbd_lineage_entry):
            parent_lineage_id = int(pbd_lineage_entry[1])
            if parent_lineage_id < 0:
                parent_lineage_id = abs(parent_lineage_id)
                is_parent_orthospecies = False
            else:
                is_parent_orthospecies = True
            extinction_time = float(pbd_lineage_entry[4])
            if extinction_time < 0:
                extinction_time = None
            lineage = cls(
                    lineage_id=int(pbd_lineage_entry[0]),
                    parent_lineage_id=parent_lineage_id,
                    is_parent_orthospecies=is_parent_orthospecies,
                    origin_time=float(pbd_lineage_entry[2]),
                    speciation_completion_time=float(pbd_lineage_entry[3]),
                    extinction_time=extinction_time,
                    species_id=int(pbd_lineage_entry[5]),
                    )
            return lineage

        def __init__(self, **kwargs):
            self.lineage_id = kwargs.pop("lineage_id")
            self.parent_lineage_id = kwargs.pop("parent_lineage_id")
            self.is_parent_orthospecies = kwargs.pop("is_parent_orthospecies")
            self._origin_time = kwargs.pop("origin_time")
            self.speciation_completion_time = kwargs.pop("speciation_completion_time")
            self.extinction_time = kwargs.pop("extinction_time")
            self.species_id = kwargs.pop("species_id")
            self.label = kwargs.pop("label", None)
            self.time = kwargs.pop("time", None)
            self.lineage_node = kwargs.pop("lineage_node", None)
            self.species_node = kwargs.pop("species_node", None)

        def __lt__(self, o):
            # shouldn't matter, but we actually want younger lineages first,
            # so we sort in *reverse* lineage_id
            return self.lineage_id > o.lineage_id

        def clone(self):
            return self.__class__(
                lineage_id=self.lineage_id,
                parent_lineage_id=self.parent_lineage_id,
                is_parent_orthospecies=self.is_parent_orthospecies,
                origin_time=self._origin_time,
                speciation_completion_time=self.speciation_completion_time,
                extinction_time=self.extinction_time,
                species_id=self.species_id,
                label=self.label,
                time=self.time,
                lineage_node=self.lineage_node,
                species_node=self.species_node,
                )

        def _get_is_extinct(self):
            return not self.extinction_time is None
        is_extinct = property(_get_is_extinct)

        # Non-mutable as _LineageQueue uses this as a sort key in the heap Only
        # way to modify is to pop, create new and reinsert
        def _get_origin_time(self):
            return self._origin_time
        origin_time = property(_get_origin_time)

    class _LineageQueue(object):

        class LineageQueueEmptyException(Exception):
            pass

        def __init__(self):
            self._lineages = []
            heapq.heapify(self._lineages)
            self._lineage_id_lineage_entry_map = {}
            self._lineage_id_to_original_lineage_map = {}

        def new_lineage(self, **kwargs):
            lineage = ProtractedSpeciationProcess._Lineage(**kwargs)
            self.push_lineage(lineage, is_copy=False)
            return lineage

        def push_lineage(self, lineage, is_copy=True):
            # assert lineage.lineage_id not in self._lineage_id_lineage_entry_map
            # assert lineage.lineage_id not in self._lineage_id_to_original_lineage_map
            if is_copy:
                stored_lineage = lineage.clone()
            else:
                stored_lineage = lineage
            self._lineage_id_to_original_lineage_map[lineage.lineage_id] = lineage
            self._lineage_id_lineage_entry_map[stored_lineage.lineage_id] = stored_lineage
            heapq.heappush(self._lineages, (-stored_lineage.origin_time, stored_lineage))

        def register_lineage_reference(self, lineage):
            self._lineage_id_to_original_lineage_map[lineage.lineage_id] = lineage

        def has_lineage(self, lineage):
            return lineage.lineage_id in self._lineage_id_lineage_entry_map

        def get_lineage(self, lineage_id):
            return self._lineage_id_lineage_entry_map[lineage_id]

        def get_lineage_by_index(self, idx):
            return self._lineages[0][1]

        def get_original_lineage(self, lineage_id):
            return self._lineage_id_to_original_lineage_map[lineage_id]

        def pop_youngest_lineage(self):
            # try:
            #     key, lineage = heapq.heappop(self._lineages)
            # except IndexError:
            #     raise ProtractedSpeciationProcess._LineageQueue.LineageQueueEmptyException from IndexError
            if not self._lineages:
                raise ProtractedSpeciationProcess._LineageQueue.LineageQueueEmptyException
            key, lineage = heapq.heappop(self._lineages)
            # del self._lineage_id_lineage_entry_map[lineage.lineage_id]
            # del self._lineage_id_to_original_lineage_map[lineage.lineage_id]
            return lineage

        def __len__(self):
            return len(self._lineages)

        def get_initial_lineage(self):
            return self._lineages[0][1]

    def __init__(self,
            speciation_initiation_from_orthospecies_rate,
            speciation_initiation_from_incipient_species_rate,
            speciation_completion_rate,
            orthospecies_extinction_rate,
            incipient_species_extinction_rate,
            lineage_label_format_template=None,
            species_label_format_template=None,
            rng=None,
            **kwargs
            ):
        self.speciation_initiation_from_orthospecies_rate = speciation_initiation_from_orthospecies_rate
        self.orthospecies_extinction_rate = orthospecies_extinction_rate
        self.speciation_initiation_from_incipient_species_rate = speciation_initiation_from_incipient_species_rate
        self.speciation_completion_rate = speciation_completion_rate
        self.incipient_species_extinction_rate = incipient_species_extinction_rate
        self.is_initial_lineage_orthospecies = kwargs.get("is_initial_lineage_orthospecies", False)
        self.species_lineage_sampling_scheme = kwargs.get("species_lineage_sampling_scheme", "random") # 'random', 'oldest', 'youngest'
        if lineage_label_format_template is None:
            self.lineage_label_format_template = "S{species_id}.L{lineage_id}"
        else:
            self.lineage_label_format_template = lineage_label_format_template
        if species_label_format_template is None:
            self.species_label_format_template = "S{species_id}"
        else:
            self.species_label_format_template = species_label_format_template
        if rng is None:
            self.rng = GLOBAL_RNG
        else:
            self.rng = rng
        self.lineage_tree_to_species_tree_node_attr = "species_tree_node"
        self.species_tree_to_lineage_tree_node_attr = "lineage_tree_nodes"

    def generate_sample(self, **kwargs):
        """

        Samples from the Protracted Speciation Model process, returning a tuple:

            -   the lineage tree: this tree has all nodes/lineages, i.e. both
                "good" species as well as incipient species.
            -   the (ortho- or confirmed- or "good"-)species tree: the tree
                only has "good" species, i.e. with all incipient species
                pruned out.

        Each tip node on the lineage tree will have an attribute,
        ``species_tree_node``, which is a reference to the species tree tip to
        which the lineage tip corresponds.

        Each tip node on the species tree will have an attribute,
        ``lineage_tree_nodes``, which is a reference to an iterable of
        tip nodes on the lineage tree that are associated with
        this species tree node.

        Parameters
        ----------

        max_time : float or |None|
            Terminate and return results if this time is exceeded. If |None|,
            then do not terminate based on run time.
        min_extant_orthospecies : int or |None|
            Terminate and return results when at least this number of tips are
            found in the confirmed-species tree (i.e., the pruned tree
            consisting of only "good" species) and other specified conditions
            are met.
        max_extant_orthospecies : int or |None|
            Terminate and return results when at least this number of tips are
            found in the confirmed-species tree (i.e., the pruned tree
            consisting of only "good" species).
        num_extant_lineages : int or |None|
            Terminate and return results when this number of tips are found in
            the lineage tree (i.e. the tree with both incipient and good
            species). If |None|, then do not terminate based on the
            number of tipes on the incipient species tree.
        is_retry_on_total_extinction : bool
            If |False|, then a TreeSimTotalExtinctionException will be raised
            if all lineages go extinct before the termination conditions are
            met. Defaults to |True|: if all lineages go extinct before the
            termination conditions are met, then the simulation is rerun, up to
            a maximum of ``max_retries``.
        max_retries : int
            Maximum number of runs to execute in the event of
            prematurely-terminated simulations due to all lineages going
            extinct. Once this number or re-runs is exceed, then
            TreeSimTotalExtinctionException is raised. Defaults to 1000. Set to
            |None| to never quit trying.

        Returns
        -------
        lineage_tree : |Tree| instance
            A tree from the protracted speciation process, with all lineages
            (good species as well as incipient species).
        orthospecies_tree : |Tree| instance
            A tree from the protracted speciation process with only "good" species.

        """
        is_retry_on_total_extinction = kwargs.pop("is_retry_on_total_extinction", True)
        max_retries = kwargs.pop("max_retries", 1000)
        num_retries = 0
        lineage_tree = None
        orthospecies_tree = None
        while True:
            try:
                lineage_tree, orthospecies_tree = self._generate_trees(**kwargs)
                break
            except ProcessFailedException:
                if not is_retry_on_total_extinction:
                    raise
                num_retries += 1
                if max_retries is not None and num_retries > max_retries:
                    raise
        assert lineage_tree is not None
        return lineage_tree, orthospecies_tree

    def _generate_trees(self, **kwargs):
        max_time = kwargs.get("max_time", None)
        num_extant_lineages = kwargs.get("num_extant_lineages", None)
        min_extant_lineages = kwargs.get("min_extant_lineages", None)
        max_extant_lineages = kwargs.get("max_extant_lineages", None)
        num_extant_orthospecies = kwargs.get("num_extant_orthospecies", None)
        min_extant_orthospecies = kwargs.get("min_extant_orthospecies", None)
        max_extant_orthospecies = kwargs.get("max_extant_orthospecies", None)
        lineage_taxon_namespace = kwargs.get("lineage_taxon_namespace", None)
        species_taxon_namespace = kwargs.get("species_taxon_namespace", None)
        lineage_data = []
        for idx in range(2):
            lineage_data.append({
                "lineage_id": 0,
                "species_id": 0,
                "lineage_collection": []
                })
        phase_idx = 0
        while phase_idx < 2:
            # Run two passes if 'max_time' specified, with each pass building
            # up one side of the root and only accepted if there are at least
            # one surviving lineage; this way, condition crown age on
            # 'max_time'.
            self._generate_lineages(
                    lineage_data=lineage_data,
                    max_time=max_time,
                    num_extant_lineages=num_extant_lineages,
                    min_extant_lineages=min_extant_lineages,
                    max_extant_lineages=max_extant_lineages,
                    num_extant_orthospecies=num_extant_orthospecies,
                    min_extant_orthospecies=min_extant_orthospecies,
                    max_extant_orthospecies=max_extant_orthospecies,
                    phase_idx=phase_idx,
                    lineage_taxon_namespace=lineage_taxon_namespace,
                    species_taxon_namespace=species_taxon_namespace,
                    )
            if num_extant_orthospecies is not None or max_extant_orthospecies is not None or min_extant_orthospecies is not None:
                if "lineage_tree" in lineage_data[phase_idx]:
                    return lineage_data[phase_idx]["lineage_tree"], lineage_data[phase_idx]["orthospecies_tree"]
                else:
                    raise ProcessFailedException()
            elif num_extant_lineages is not None or max_extant_lineages is not None or min_extant_lineages is not None:
                break
            elif phase_idx == 0 and (len(lineage_data[0]["orthospecies_lineages"]) + len(lineage_data[0]["incipient_species_lineages"]) > 0):
                phase_idx += 1
            elif phase_idx == 1 and self._check_good(
                    orthospecies_lineages=lineage_data[1]["orthospecies_lineages"],
                    incipient_species_lineages=lineage_data[1]["incipient_species_lineages"]):
                phase_idx += 1
        # lineage_collection = itertools.chain(
        #         lineage_data[0]["lineage_collection"],
        #         lineage_data[1]["lineage_collection"],
        #         )
        lineage_collection = lineage_data[0]["lineage_collection"] + lineage_data[1]["lineage_collection"]
        lineage_tree = self._compile_lineage_tree(
                lineage_collection=lineage_collection,
                max_time=max_time if max_time is not None else lineage_data[phase_idx]["final_time"],
                is_drop_extinct=True,
                )
        orthospecies_tree = self._compile_species_tree(
                lineage_collection=lineage_collection,
                max_time=max_time if max_time is not None else lineage_data[phase_idx]["final_time"],
                )
        return self._finalize_trees(
                lineage_tree=lineage_tree,
                lineage_taxon_namespace=lineage_taxon_namespace,
                orthospecies_tree=orthospecies_tree,
                species_taxon_namespace=species_taxon_namespace,
                lineage_collection=lineage_collection,
                )

    def _generate_lineages(self, **kwargs):
        current_time = 0.0
        lineage_data = kwargs.get("lineage_data")
        max_time = kwargs.get("max_time", None)
        num_extant_lineages = kwargs.get("num_extant_lineages", None)
        min_extant_lineages = kwargs.get("min_extant_lineages", None)
        max_extant_lineages = kwargs.get("max_extant_lineages", None)
        num_extant_orthospecies = kwargs.get("num_extant_orthospecies", None)
        min_extant_orthospecies = kwargs.get("min_extant_orthospecies", None)
        max_extant_orthospecies = kwargs.get("max_extant_orthospecies", None)
        phase_idx = kwargs.get("phase_idx")
        lineage_taxon_namespace = kwargs.get("lineage_taxon_namespace", None)
        species_taxon_namespace = kwargs.get("species_taxon_namespace", None)
        if phase_idx == 0:
            lineage_data[phase_idx]["lineage_id"] = 1
            lineage_data[phase_idx]["species_id"] = 1
            initial_lineage = self._new_lineage(
                    lineage_id=lineage_data[phase_idx]["lineage_id"],
                    parent_lineage=None,
                    origin_time=-1e-10,
                    )
            lineage_data[phase_idx]["orthospecies_lineages"] = [initial_lineage]
            lineage_data[phase_idx]["incipient_species_lineages"] = []
        else:
            lineage_data[phase_idx]["lineage_id"] = lineage_data[0].get("lineage_id", 0) + 1
            lineage_data[phase_idx]["species_id"] = lineage_data[0].get("species_id", 0)
            initial_lineage = self._new_lineage(
                    lineage_id=lineage_data[phase_idx]["lineage_id"],
                    parent_lineage=lineage_data[0]["lineage_collection"][0],
                    origin_time=current_time,
                    )
            lineage_data[phase_idx]["orthospecies_lineages"] = []
            lineage_data[phase_idx]["incipient_species_lineages"] = [initial_lineage]
        lineage_data[phase_idx]["lineage_collection"] = [initial_lineage]
        lineage_collection = lineage_data[phase_idx]["lineage_collection"]
        orthospecies_lineages = lineage_data[phase_idx]["orthospecies_lineages"]
        incipient_species_lineages = lineage_data[phase_idx]["incipient_species_lineages"]
        while True:
            num_orthospecies = len(orthospecies_lineages)
            num_incipient_species = len(incipient_species_lineages)
            if num_incipient_species + num_orthospecies == 0:
                raise TreeSimTotalExtinctionException()
            ## Draw time to next event
            event_rates = []
            # Event type 0
            event_rates.append(self.speciation_initiation_from_orthospecies_rate * num_orthospecies)
            # Event type 1
            event_rates.append(self.orthospecies_extinction_rate * num_orthospecies)
            # Event type 2
            event_rates.append(self.speciation_initiation_from_incipient_species_rate * num_incipient_species)
            # Event type 3
            event_rates.append(self.speciation_completion_rate * num_incipient_species)
            # Event type 4
            event_rates.append(self.incipient_species_extinction_rate * num_incipient_species)
            # All events
            rate_of_any_event = sum(event_rates)
            # Waiting time
            waiting_time = self.rng.expovariate(rate_of_any_event)
            # waiting_time = -math.log(self.rng.uniform(0, 1))/rate_of_any_event
            if max_time and (current_time + waiting_time) > max_time:
                current_time = max_time
                break
            # we do this here so that the (newest) tip lineages have the
            # waiting time to the next event branch lengths
            if (num_extant_lineages is not None
                    or min_extant_lineages is not None
                    or max_extant_lineages is not None):
                has_lineage_count_requirements = True
                if (
                        (num_extant_lineages is None or ((num_incipient_species + num_orthospecies) == num_extant_lineages))
                        and (min_extant_lineages is None or ((num_incipient_species + num_orthospecies) >= min_extant_lineages))
                        and (max_extant_lineages is None or ((num_incipient_species + num_orthospecies) == max_extant_lineages))
                        ):
                    is_lineage_count_requirements_met = True
                else:
                    is_lineage_count_requirements_met = False
            else:
                has_lineage_count_requirements = False
                is_lineage_count_requirements_met = None
            if max_extant_lineages is not None and (num_incipient_species + num_orthospecies) > max_extant_lineages:
                raise ProcessFailedException()
            if num_extant_orthospecies is not None or max_extant_orthospecies is not None or min_extant_orthospecies is not None:
                ## note: very expensive operation to count orthospecies leaves!
                has_orthospecies_count_requirements = True
                is_orthospecies_count_requirements_met = False
                final_time = current_time + self.rng.uniform(0, waiting_time)
                lineage_collection_snapshot = [lineage.clone() for lineage in itertools.chain(lineage_data[0]["lineage_collection"], lineage_data[1]["lineage_collection"])]
                try:
                    orthospecies_tree = self._compile_species_tree(
                            lineage_collection=lineage_collection_snapshot,
                            max_time=final_time,
                            )
                    num_leaves = len(orthospecies_tree.leaf_nodes())
                    if (
                            (num_extant_orthospecies is None or num_leaves == num_extant_orthospecies)
                            and (min_extant_orthospecies is None or num_leaves >= min_extant_orthospecies)
                            and (max_extant_orthospecies is None or num_leaves <= max_extant_orthospecies)
                            ):
                        lineage_tree = self._compile_lineage_tree(
                            lineage_collection=lineage_collection_snapshot,
                            max_time=final_time,
                            is_drop_extinct=True,
                            )
                        lineage_tree, orthospecies_tree = self._finalize_trees(
                                lineage_tree=lineage_tree,
                                lineage_taxon_namespace=lineage_taxon_namespace,
                                orthospecies_tree=orthospecies_tree,
                                species_taxon_namespace=species_taxon_namespace,
                                lineage_collection=lineage_collection_snapshot,
                                )
                        lineage_data[phase_idx]["lineage_tree"] = lineage_tree
                        lineage_data[phase_idx]["orthospecies_tree"] = orthospecies_tree
                        is_orthospecies_count_requirements_met = True
                except ProcessFailedException:
                    pass
                if max_extant_orthospecies is not None and num_leaves > max_extant_orthospecies:
                    raise ProcessFailedException
            else:
                has_orthospecies_count_requirements = False
                is_orthospecies_count_requirements_met = None
            if (
                    ( (has_lineage_count_requirements and is_lineage_count_requirements_met) and (has_orthospecies_count_requirements and is_orthospecies_count_requirements_met) )
                    or ( (has_lineage_count_requirements and is_lineage_count_requirements_met) and (not has_orthospecies_count_requirements) )
                    or ( (not has_lineage_count_requirements) and (has_orthospecies_count_requirements and is_orthospecies_count_requirements_met) )
            ):
                final_time = current_time + self.rng.uniform(0, waiting_time)
                lineage_data[phase_idx]["final_time"] = final_time
                break
            else:
                # add to current time
                current_time += waiting_time
                # Select event
                event_type_idx = probability.weighted_index_choice(weights=event_rates, rng=self.rng)
                assert (event_type_idx >= 0 and event_type_idx <= 4)
                if event_type_idx == 0:
                    # Splitting of new incipient species lineage from orthospecies lineage
                    parent_lineage = self.rng.choice(orthospecies_lineages)
                    lineage_data[phase_idx]["lineage_id"] += 1
                    new_lineage = self._new_lineage(
                            lineage_id=lineage_data[phase_idx]["lineage_id"],
                            parent_lineage=parent_lineage,
                            origin_time=current_time,
                            )
                    lineage_collection.append(new_lineage)
                    incipient_species_lineages.append(new_lineage)
                elif event_type_idx == 1:
                    # Extinction of an orthospecies lineage
                    lineage_idx = self.rng.randint(0, len(orthospecies_lineages)-1)
                    orthospecies_lineages[lineage_idx].extinction_time = current_time
                    del orthospecies_lineages[lineage_idx]
                elif event_type_idx == 2:
                    # Splitting of new incipient species lineage from incipient lineage
                    parent_lineage = self.rng.choice(incipient_species_lineages)
                    lineage_data[phase_idx]["lineage_id"] += 1
                    new_lineage = self._new_lineage(
                            lineage_id=lineage_data[phase_idx]["lineage_id"],
                            parent_lineage=parent_lineage,
                            origin_time=current_time,
                            )
                    lineage_collection.append(new_lineage)
                    incipient_species_lineages.append(new_lineage)
                elif event_type_idx == 3:
                    # Completion of speciation
                    lineage_idx = self.rng.randint(0, len(incipient_species_lineages)-1)
                    lineage = incipient_species_lineages[lineage_idx]
                    lineage.speciation_completion_time = current_time
                    lineage_data[phase_idx]["species_id"] += 1
                    lineage.species_id = lineage_data[phase_idx]["species_id"]
                    orthospecies_lineages.append(lineage)
                    del incipient_species_lineages[lineage_idx]
                elif event_type_idx == 4:
                    # Extinction of an incipient_species lineage
                    lineage_idx = self.rng.randint(0, len(incipient_species_lineages)-1)
                    incipient_species_lineages[lineage_idx].extinction_time = current_time
                    del incipient_species_lineages[lineage_idx]
                else:
                    raise Exception("Unexpected event type index: {}".format(event_type_idx))

    def _new_lineage(self,
            lineage_id,
            parent_lineage,
            origin_time=None,
            ):
        if parent_lineage is None:
            parent_lineage_id = 0
            species_id = 1
            is_parent_orthospecies = None
        else:
            parent_lineage_id = parent_lineage.lineage_id
            species_id = parent_lineage.species_id
            is_parent_orthospecies = parent_lineage.speciation_completion_time is not None
        lineage = ProtractedSpeciationProcess._Lineage(
                lineage_id=lineage_id,
                parent_lineage_id=parent_lineage_id,
                is_parent_orthospecies=is_parent_orthospecies,
                origin_time=origin_time,
                speciation_completion_time=None,
                extinction_time=None,
                species_id=species_id,
                )
        lineage._check_parent_lineage = parent_lineage # for _check_good
        return lineage

    def _correlate_lineage_and_species_trees(self, lineage_collection):
        seen_lineage_nodes = set()
        seen_species_nodes = set()
        lineage_id_to_species_maps = {}
        species_id_lineage_node_collection_map = {}
        species_id_species_node_map = {}
        for lineage in lineage_collection:
            if not lineage.lineage_node:
                assert not lineage.species_node
                continue
            if lineage.lineage_node.taxon is None:
                continue
            lineage_node = lineage.lineage_node
            species_node = lineage.species_node
            assert lineage_node not in seen_lineage_nodes
            seen_lineage_nodes.add(lineage_node)
            try:
                species_id_lineage_node_collection_map[lineage_node._species_id].add(lineage_node)
            except KeyError:
                species_id_lineage_node_collection_map[lineage_node._species_id] = set([lineage_node])
            if species_node is not None:
                assert species_node not in seen_species_nodes
                seen_species_nodes.add(lineage.species_node)
                species_id_species_node_map[species_node._species_id] = species_node
                setattr(species_node, self.species_tree_to_lineage_tree_node_attr, species_id_lineage_node_collection_map[species_node._species_id])
        for species_id, lineage_nodes in species_id_lineage_node_collection_map.items():
            for nd in lineage_nodes:
                setattr(nd, self.lineage_tree_to_species_tree_node_attr, species_id_species_node_map[species_id])

    def _finalize_trees(self,
            lineage_tree,
            lineage_taxon_namespace,
            orthospecies_tree,
            species_taxon_namespace,
            lineage_collection,
            ):
        self._build_taxa(tree=lineage_tree, taxon_namespace=lineage_taxon_namespace)
        self._build_taxa(tree=orthospecies_tree, taxon_namespace=species_taxon_namespace)
        self._correlate_lineage_and_species_trees(lineage_collection=lineage_collection)
        return lineage_tree, orthospecies_tree

    def _compile_species_tree(self,
            lineage_collection,
            max_time,
            ):
        if self.species_lineage_sampling_scheme == "oldest":
            lt = sorted(lineage_collection, key=lambda x: x.origin_time)
        elif self.species_lineage_sampling_scheme == "youngest":
            lt = sorted(lineage_collection, key=lambda x: x.origin_time, reverse=True)
        elif self.species_lineage_sampling_scheme == "random":
            lt = self.rng.sample(lineage_collection, len(lineage_collection))
        else:
            raise ValueError(self.species_lineage_sampling_scheme)
        seen_species_ids = set()
        to_restore_species_extinction_times = {}
        for lineage_entry in lt:
            if lineage_entry.extinction_time is None:
                if lineage_entry.species_id not in seen_species_ids:
                    seen_species_ids.add(lineage_entry.species_id)
                else:
                    to_restore_species_extinction_times[lineage_entry] = lineage_entry.extinction_time
                    lineage_entry.extinction_time = max_time # pseudo-extinction
        t = self._compile_tree(
                lineage_collection=lt,
                max_time=max_time,
                tree_type="species",
                is_drop_extinct=True,
                )
        for k,v in to_restore_species_extinction_times.items():
            k.extinction_time = v
        return t

    def _compile_lineage_tree(self,
            lineage_collection,
            max_time,
            is_drop_extinct=True,
            ):
        return self._compile_tree(
            lineage_collection=lineage_collection,
            max_time=max_time,
            tree_type="lineage",
            is_drop_extinct=is_drop_extinct,
            )

    def _compile_tree(self,
            lineage_collection,
            max_time,
            tree_type,
            is_drop_extinct=True,
            ):
        if tree_type == "lineage":
            label_template = self.lineage_label_format_template
            node_attr = "lineage_node"
        elif tree_type == "species":
            label_template = self.species_label_format_template
            node_attr = "species_node"
        else:
            raise ValueError(tree_type)
        if len(lineage_collection) == 1:
            tree = dendropy.Tree(is_rooted=True)
            label = label_template.format(species_id=1, lineage_id=0)
            tree.seed_node._taxon_label = label
            tree.seed_node._lineage_id = 0
            tree.seed_node._species_id = 1
            tree.seed_node.edge.length = max_time
            tree.seed_node._time = max_time
            setattr(lineage_collection[0], node_attr, tree.seed_node)
            return tree
        lineage_queue = self._build_lineage_queue(
                lineage_collection=lineage_collection,
                max_time=max_time,
                is_drop_extinct=is_drop_extinct,
                node_attr=node_attr,
                label_template=label_template,
                )
        while True:
            daughter_lineage = lineage_queue.pop_youngest_lineage()
            parent_lineage_id = daughter_lineage.parent_lineage_id
            try:
                parent_lineage = lineage_queue.get_lineage(parent_lineage_id)
                start_time = daughter_lineage.origin_time
                sp_comp_time = daughter_lineage.speciation_completion_time
                daughter_node = getattr(daughter_lineage, node_attr)
                end_time = daughter_node._time
                parent_sp_comp_time = parent_lineage.speciation_completion_time
                parent_node = getattr(parent_lineage, node_attr)
                parent_end_time = parent_node._time
                ch1 = parent_node
                ch1.edge.length = parent_end_time - start_time
                ch2 = daughter_node
                ch2.edge.length = end_time - start_time
                new_node = dendropy.Node()
                new_node.add_child(ch1)
                new_node.add_child(ch2)
                new_node._time = daughter_lineage.origin_time
                setattr(parent_lineage, node_attr, new_node)
            except KeyError:
                if parent_lineage_id != 0:
                    parent_lineage = lineage_queue.get_original_lineage(parent_lineage_id) #lineage_id_lineage_entry_idx_map[parent_lineage_id]
                    parent_lineage_clone = parent_lineage.clone()
                    setattr(parent_lineage_clone, node_attr, getattr(daughter_lineage, node_attr))
                    parent_sp_comp_time2 = parent_lineage.speciation_completion_time
                    if not (parent_sp_comp_time2 is not None and (parent_sp_comp_time2 < daughter_lineage.origin_time)):
                        parent_lineage_clone.speciation_completion_time = daughter_lineage.speciation_completion_time
                    lineage_queue.push_lineage(parent_lineage_clone, is_copy=False)
            if len(lineage_queue) < 2:
                if len(lineage_queue) == 0:
                    raise ProcessFailedException
                elif len(lineage_queue) == 1:
                    initial_lineage = lineage_queue.get_initial_lineage()
                else:
                    initial_lineage = None
                    while True:
                        try:
                            initial_lineage = lineage_queue.pop_youngest_lineage()
                        except ProtractedSpeciationProcess._LineageQueue.LineageQueueEmptyException as e:
                            break
                    assert initial_lineage is not None
                seed_node = getattr(initial_lineage, node_attr)
                seed_node.edge.length = initial_lineage.origin_time
                tree = dendropy.Tree(
                        seed_node=seed_node,
                        is_rooted=True,
                        )
                return tree
            if parent_lineage_id == 0:
                raise ValueError

    def _build_taxa(self, tree, taxon_namespace):
        if taxon_namespace is None:
            taxon_namespace = dendropy.TaxonNamespace()
        assert len(tree.taxon_namespace) == 0
        tree.taxon_namespace = taxon_namespace
        for nd in tree.leaf_node_iter():
            nd.taxon = tree.taxon_namespace.require_taxon(label=nd._taxon_label)
            del nd._taxon_label
        return tree

    def _build_lineage_queue(self,
            lineage_collection,
            max_time,
            is_drop_extinct,
            node_attr,
            label_template,
            ):
        lineageq = ProtractedSpeciationProcess._LineageQueue()
        for lineage in lineage_collection:
            if not is_drop_extinct or not lineage.is_extinct:
                node = dendropy.Node()
                if is_drop_extinct:
                    node._time = max_time
                else:
                    node._time = lineage.extinction_time if lineage.extinction_time is not None else max_time
                label = label_template.format(species_id=lineage.species_id, lineage_id=lineage.lineage_id)
                node._taxon_label = label
                node._lineage_id = lineage.lineage_id
                node._species_id = lineage.species_id
                setattr(lineage, node_attr, node)
                lineageq.push_lineage(lineage=lineage, is_copy=True)
            else:
                lineageq.register_lineage_reference(lineage=lineage)
        return lineageq

    def _check_good(self, orthospecies_lineages, incipient_species_lineages):
        if orthospecies_lineages:
            return True
        if not incipient_species_lineages:
            return False
        for lineage in incipient_species_lineages:
            parent_lineage = lineage._check_parent_lineage
            origin_time = lineage.origin_time
            while parent_lineage is not None and parent_lineage.lineage_id > 1:
                if parent_lineage.speciation_completion_time is not None and parent_lineage.speciation_completion_time < origin_time:
                    return True
                origin_time = parent_lineage.origin_time
                parent_lineage = parent_lineage._check_parent_lineage
        return False



#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Various text-manipulating and formatting utilities.
"""

import re
import sys
import time
import itertools
import locale
import codecs

###############################################################################
## Cross-version compatibility
try:
    from StringIO import StringIO # Python 2 legacy support: StringIO in this module is the one needed (not io)
except ImportError:
    from io import StringIO # Python 3

###############################################################################
## Unicode/String Conversions

try:
    ENCODING = locale.getencoding()
except:
    try:
        ENCODING = locale.getdefaultlocale()[1]
    except ValueError:
        ENCODING = None # let default value be assigned below

if ENCODING == None:
    ENCODING = 'UTF-8'

def bytes_to_text(s):
    """
    Converts a byte string (as read from, e.g., standard input)
    to a text string.

    In Python 3, this is from type ``bytes`` to ``str``.
    In Python 2, this is, confusingly, from type ``str`` to ``unicode``.

    """
    s = codecs.decode(s, ENCODING)
    if sys.hexversion < 0x03000000:
        s = codecs.encode(s, "utf-8")
    return s

def parse_curie_standard_qualified_name(prefixed_name, sep=":"):
    if sep not in prefixed_name:
        raise ValueError("'{}' is not a valid CURIE-standard qualified name".format(prefixed_name))
    return prefixed_name.split(":", 1)

## From:
    # The Peyotl module of the Open Tree of Life Project
    # Mark T. Holder
    # https://github.com/mtholder/peyotl
    # https://github.com/mtholder/peyotl/blob/c3a544211edc669e664bae28095d52cecfa004f3/peyotl/utility/str_util.py#L5-L25
if sys.version_info.major == 2:
    import __builtin__ as builtins  # extra verbosity to mollify linter
    def is_str_type(x):
        return isinstance(x, builtins.basestring)
else:
    def is_str_type(x):
        return isinstance(x, str)

###############################################################################
##

def camel_case(s):
    components = s.split('_')
    return components[0] + "".join(x.title() for x in components[1:])

def snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()

###############################################################################
##

def unique_taxon_label_map(taxa, taxon_label_map=None, max_label_len=0, logger=None):
    """
    Given a list of taxa, returns a dictionary with the Taxon objects as
    keys and string labels as values, where the labels are guaranteed to
    be unique. If ``taxon_label_map`` is pre-populated (as <Taxon> : 'label'),
    then those labels will be used as the basis for the label composition,
    otherwise the original taxon object label will be used. ``max_label_len``
    can be used to restrict the maximum length of the labels.
    """
    if taxon_label_map is None:
        taxon_label_map = {}
        for t in taxa:
            taxon_label_map[t] = t.label
    labels = []
    for t in taxon_label_map:
        label = taxon_label_map[t]
        idx = 1
        if label in labels:
            candidate_label = label
            while candidate_label in labels:
                idx += 1
                if max_label_len > 0:
                    k = max_label_len - len(str(idx))
                    if k < 1:
                        raise ValueError("Unable to make labels unique with maximum label length of %d" % max_label_len)
                    candidate_label = label[:k] + str(idx)
                else:
                    candidate_label = label + str(idx)
            label = candidate_label
        labels.append(label)
        taxon_label_map[t] = label
    return taxon_label_map

###############################################################################
##

def format_dict_table(*args, **kwargs):
    """
    Returns a (single) string representation of a tuple of dictionaries in a
    table format. This method can read the column names directly off the
    dictionary keys, but if a tuple of these keys is provided in the
    'column_names' variable, then the order of column_names will follow the
    order of the fields/keys in that variable.
    """
    display = format_dict_table_rows(*args, **kwargs)
    if display:
        return "\n".join(display)
    else:
        return ""

def format_dict_table_rows(rows, column_names=None, max_column_width=None, border_style=2):
    """
    Returns a string representation of a tuple of dictionaries in a
    table format. This method can read the column names directly off the
    dictionary keys, but if a tuple of these keys is provided in the
    'column_names' variable, then the order of column_names will follow
    the order of the fields/keys in that variable.
    """
    if column_names or len(rows) > 0:
        lengths = {}
        rules = {}
        if column_names:
            column_list = column_names
        else:
            try:
                column_list = rows[0].keys()
            except:
                column_list = None
        if column_list:
            # characters that make up the table rules
            border_style = int(border_style)
            #border_style = 0
            if border_style == 0:
                vertical_rule = '  '
                horizontal_rule = ''
                rule_junction = ''
            elif border_style == 1:
                vertical_rule = ' '
                horizontal_rule = '-'
                rule_junction = '-'
            else:
                vertical_rule = ' | '
                horizontal_rule = '-'
                rule_junction = '-+-'
            if border_style >= 3:
                left_table_edge_rule = '| '
                right_table_edge_rule = ' |'
                left_table_edge_rule_junction = '+-'
                right_table_edge_rule_junction = '-+'
            else:
                left_table_edge_rule = ''
                right_table_edge_rule = ''
                left_table_edge_rule_junction = ''
                right_table_edge_rule_junction = ''

            if max_column_width:
                column_list = [c[:max_column_width] for c in column_list]
                trunc_rows = []
                for row in rows:
                    new_row = {}
                    for k in row.keys():
                        new_row[k[:max_column_width]] = str(row[k])[:max_column_width]
                    trunc_rows.append(new_row)
                rows = trunc_rows

            for col in column_list:
                rls = [len(str(row[col])) for row in rows]
                lengths[col] = max(rls+[len(col)])
                rules[col] = horizontal_rule*lengths[col]

            template_elements = ["%%(%s)-%ss" % (col, lengths[col]) for col in column_list]
            row_template = vertical_rule.join(template_elements)
            border_template = rule_junction.join(template_elements)
            full_line = left_table_edge_rule_junction + (border_template % rules) + right_table_edge_rule_junction
            display = []
            if border_style > 0:
                display.append(full_line)
            display.append(left_table_edge_rule + (row_template % dict(zip(column_list, column_list))) + right_table_edge_rule)
            if border_style > 0:
                display.append(full_line)
            for row in rows:
                display.append(left_table_edge_rule + (row_template % row) + right_table_edge_rule)
            if border_style > 0:
                display.append(full_line)
            return display
        else:
            return ''
    else:
        return ''



#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
DEPRECATED IN DENDROPY 4: USE `dendropy.model.coalescent` instead.
"""

from dendropy.model import coalescent
from dendropy.utility import deprecate
from dendropy.utility import constants

def discrete_time_to_coalescence(n_genes,
                                 pop_size=None,
                                 n_to_coalesce=2,
                                 rng=None):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.coalescent' module has moved to 'dendropy.model.coalescent'.",
            old_construct="from dendropy import coalescent\ncoalescent.discrete_time_to_coalescence(...)",
            new_construct="from dendropy.model import coalescent\ncoalescent.discrete_time_to_coalescence(...)")
    return coalescent.discrete_time_to_coalescence(
            n_genes=n_genes,
            pop_size=pop_size,
            n_to_coalesce=n_to_coalesce,
            rng=rng)

def time_to_coalescence(n_genes,
        pop_size=None,
        n_to_coalesce=2,
        rng=None):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.coalescent' module has moved to 'dendropy.model.coalescent'.",
            old_construct="from dendropy import coalescent\ncoalescent.time_to_coalescence(...)",
            new_construct="from dendropy.model import coalescent\ncoalescent.time_to_coalescence(...)")
    return coalescent.time_to_coalescence(
        n_genes=n_genes,
        pop_size=pop_size,
        n_to_coalesce=n_to_coalesce,
        rng=rng)

def expected_tmrca(n_genes, pop_size=None, n_to_coalesce=2):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.coalescent' module has moved to 'dendropy.model.coalescent'.",
            old_construct="from dendropy import coalescent\ncoalescent.expected_tmrca(...)",
            new_construct="from dendropy.model import coalescent\ncoalescent.expected_tmrca(...)")
    return coalescent.expected_tmrca(n_genes, pop_size=pop_size, n_to_coalesce=n_to_coalesce)

def coalesce(nodes,
             pop_size=None,
             period=None,
             rng=None,
             use_expected_tmrca=False):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.coalescent' module has moved to 'dendropy.model.coalescent', and this function has been renamed 'coalesce_nodes'.",
            old_construct="from dendropy import coalescent\ncoalescent.coalesce(...)",
            new_construct="from dendropy.model import coalescent\ncoalescent.coalesce_nodes(...)")
    return coalescent.coalesce_nodes(
            nodes=nodes,
            pop_size=pop_size,
            period=period,
            rng=rng,
            use_expected_tmrca=use_expected_tmrca)

def node_waiting_time_pairs(tree, ultrametricity_precision=constants.DEFAULT_ULTRAMETRICITY_PRECISION):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.coalescent' module has moved to 'dendropy.model.coalescent'.",
            old_construct="from dendropy import coalescent\ncoalescent.node_waiting_time_pairs(...)",
            new_construct="from dendropy.model import coalescent\ncoalescent.node_waiting_time_pairs(...)")
    return coalescent.node_waiting_time_pairs(tree=tree,
            ultrametricity_precision=ultrametricity_precision)

def extract_coalescent_frames(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.coalescent' module has moved to 'dendropy.model.coalescent'.",
            old_construct="from dendropy import coalescent\ncoalescent.extract_coalescent_frames(...)",
            new_construct="from dendropy.model import coalescent\ncoalescent.extract_coalescent_frames(...)")
    return coalescent.extract_coalescent_frames(*args, **kwargs)

def log_probability_of_coalescent_frames(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.coalescent' module has moved to 'dendropy.model.coalescent'.",
            old_construct="from dendropy import coalescent\ncoalescent.log_probability_of_coalescent_frames(...)",
            new_construct="from dendropy.model import coalescent\ncoalescent.log_probability_of_coalescent_frames(...)")
    return log_probability_of_coalescent_frames(*args, **kwargs)

def log_probability_of_coalescent_tree(*args, **kwargs):
    deprecate.dendropy_deprecation_warning(
            preamble="Deprecated since DendroPy 4: The 'dendropy.coalescent' module has moved to 'dendropy.model.coalescent'.",
            old_construct="from dendropy import coalescent\ncoalescent.log_probability_of_coalescent_tree(...)",
            new_construct="from dendropy.model import coalescent\ncoalescent.log_probability_of_coalescent_tree(...)")
    return log_probability_of_coalescent_tree(*args, **kwargs)




#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Wraps up source version control system system information.
"""

import os
import sys
import subprocess
import datetime
from dendropy.utility import textprocessing
from dendropy.utility import processio

class Revision(object):
    """
    Provides (Git) version control information
    about a project.
    """

    class VcsUnavailable(Exception):
        def __init__(self, *args, **kwargs):
            Exception.__init__(self, *args, **kwargs)

    class NonRepositoryError(Exception):
        def __init__(self, *args, **kwargs):
            Exception.__init__(self, *args, **kwargs)

    class NonBranchException(Exception):
        def __init__(self, *args, **kwargs):
            Exception.__init__(self, *args, **kwargs)

    class UntaggedException(Exception):
        def __init__(self, *args, **kwargs):
            Exception.__init__(self, *args, **kwargs)

    def __init__(self, repo_path, release=None, vcs_app_path='git'):
        self.vcs_app_path = vcs_app_path
        self.repo_path = repo_path
        self.release = release
        self._commit_id = None
        self._commit_date = None
        self._commit_tag = None
        self._branch_name = None
        self._description = None
        self._long_description = None
        self._is_available = None

    def __str__(self):
        parts = []
        if self.branch_name:
            parts.append("%s-" % self.branch_name)
        if self.commit_id:
            parts.append(self.commit_id[:10])
        if self.commit_date:
            parts.append(", %s" % str(self.commit_date))
        if parts:
            return "".join(parts)
        else:
            return ""

    def __repr__(self):
        return "<%s: '%s'>" % (self.__class__.__name__, self.__str__())

    @property
    def commit_id(self):
        if self._commit_id is None:
            self.update()
        return self._commit_id

    @property
    def commit_date(self):
        if self._commit_date is None:
            self.update()
        return self._commit_date

    @property
    def commit_tag(self):
        if self._commit_tag is None:
            self.update()
        return self._commit_tag

    @property
    def branch_name(self):
        if self._branch_name is None:
            self.update()
        return self._branch_name

    @property
    def description(self):
        if self._description is None:
            self.update()
        return self._description

    @property
    def long_description(self):
        if self._long_description is None:
            self.update()
        return self._long_description

    @property
    def is_available(self):
        if self._is_available is None:
            self.update()
        return self._is_available

    def update(self, repo_path=None):
        if repo_path is not None:
            self.repo_path = repo_path
        if not self.repo_path or not self._vcs_available():
            self._commit_id = None
            self._commit_date = None
            self._commit_tag = None
            self._branch_name = None
            self._description = None
            self._long_description = None
            self._is_available = False
            return
        self._commit_id = self.get_commit_id()
        self._commit_date = self.get_datetime()
        self._commit_tag = self.get_commit_tag()
        self._branch_name = self.get_branch_name()
        self._description = self.get_description()
        self._long_description = self._build_long_description()
        self._is_available = True

    def _run_vcs(self, cmd):
        if textprocessing.is_str_type(cmd):
            cmd = self.vcs_app_path + " " + cmd
        else:
            cmd.insert(0, self.vcs_app_path)
        try:
            p = subprocess.Popen(cmd,
                shell=True,
                cwd=os.path.abspath(self.repo_path),
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE)
            stdout, stderr = processio.communicate(p)
            retcode = p.returncode
        except OSError as e:
            return -999, "", str(e)
        return retcode, stdout, stderr

    def _vcs_available(self):
        retcode, stdout, stderr = self._run_vcs("--version")
        if retcode != 0:
            return False
        retcode, stdout, stderr = self._run_vcs("status")
        if "fatal: Not a git repository" in stderr:
            return False
        return True

    def get_commit_id(self):
        # cmd = "show --quiet --pretty=format:'%H' HEAD"
        cmd = "rev-parse --short HEAD"
        retcode, stdout, stderr = self._run_vcs(cmd)
        return stdout.replace('\n', '')

    def get_datetime(self):
        cmd = "show --quiet --pretty=format:'%at' HEAD"
        retcode, stdout, stderr = self._run_vcs(cmd)
        if stdout:
            try:
                return datetime.datetime.fromtimestamp(float(stdout.replace('\n', '').replace("'", "").replace('"','')))
            except ValueError:
                return None
        else:
            return None

    def get_commit_tag(self):
        cmd = "name-rev --name-only --tags --no-undefined HEAD"
        retcode, stdout, stderr = self._run_vcs(cmd)
        if "fatal: cannot describe" in stderr:
            return None
        else:
            return stdout.strip('\n')

    def get_branch_name(self):
        # git name-rev --name-only HEAD
        cmd = "symbolic-ref HEAD"
        retcode, stdout, stderr = self._run_vcs(cmd)
        if retcode != 0:
            if "fatal: ref HEAD is not a symbolic ref" in stderr:
                return "(no branch)"
            else:
                return None
        else:
            return stdout.replace('\n', '').split('/')[-1]

    def get_description(self):
        cmd = "describe --tags --long --always --abbrev=12"
        retcode, stdout, stderr = self._run_vcs(cmd)
        if retcode != 0:
            if "fatal: No names found, cannot describe anything." in stderr:
                return "(unnamed)"
            else:
                return None
        else:
            return stdout.replace('\n', '')

    def get_long_description(self):
        parts = []
        if self.commit_id:
            parts.append(self.commit_id)
        if self.branch_name:
            parts.append("on branch '%s'" % self.branch_name)
        if self.commit_date:
            parts.append("committed on %s" % str(self.commit_date))
        if parts:
            return ", ".join(parts)
        else:
            return None

    def _build_long_description(self):
        parts = []
        if self._commit_id:
            parts.append(self._commit_id)
        if self._branch_name:
            parts.append("on branch '%s'" % self._branch_name)
        if self._commit_date:
            parts.append("committed on %s" % str(self._commit_date))
        if parts:
            return ", ".join(parts)
        else:
            return None





#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Implementation of NEXUS-schema tree iterator.
"""

import sys
if not (sys.version_info.major >= 3 and sys.version_info.minor >= 4):
    from dendropy.utility.filesys import pre_py34_open as open
from dendropy.utility import textprocessing
from dendropy.dataio import ioservice
from dendropy.dataio import nexusreader
from dendropy.dataio import nexusprocessing

class NexusTreeDataYielder(
        ioservice.TreeDataYielder,
        nexusreader.NexusReader):

    def __init__(self,
            files=None,
            taxon_namespace=None,
            tree_type=None,
            **kwargs):
        r"""

        Parameters
        ----------
        files : iterable of sources
            Iterable of sources, which can either be strings specifying file
            paths or file-like objects open for reading. If a source element is
            a string (``isinstance(i,str) == True``), then it is assumed to be
            a path to a file. Otherwise, the source is assumed to be a file-like
            object.
        taxon_namespace : |TaxonNamespace| instance
            The operational taxonomic unit concept namespace to use to manage
            taxon definitions.
        \*\*kwargs : keyword arguments
            These will be passed directly to the base `nexusreader.NexusReader`
            class. See `nexusreader.NexusReader` for details.
        """
        ioservice.TreeDataYielder.__init__(self,
                files=files,
                taxon_namespace=taxon_namespace,
                tree_type=tree_type)
        self.assume_newick_if_not_nexus = kwargs.pop("assume_newick_if_not_nexus", False)
        kwargs["attached_taxon_namespace"] = self.attached_taxon_namespace
        nexusreader.NexusReader.__init__(self, **kwargs)
        self.exclude_chars = True
        self.exclude_trees = False

    ###########################################################################
    ## Implementation of DataYielder interface

    def _yield_items_from_stream(self, stream):
        if self._nexus_tokenizer is None:
            self.create_tokenizer(stream,
                preserve_unquoted_underscores=self.preserve_underscores)
        else:
            self._nexus_tokenizer.set_stream(stream)
        token = self._nexus_tokenizer.next_token()
        if token.upper() != "#NEXUS":
            if self.assume_newick_if_not_nexus:
                taxon_symbol_mapper = self._get_taxon_symbol_mapper(
                        taxon_namespace=self.attached_taxon_namespace,
                        enable_lookup_by_taxon_number=False,
                        )
                while True:
                    tree = self._build_tree_from_newick_tree_string(
                            tree_factory=self.tree_factory,
                            taxon_symbol_mapper=taxon_symbol_mapper)
                    if tree is None:
                        break
                    yield tree
            else:
                raise self._nexus_error("Expecting '#NEXUS', but found '{}'".format(token),
                        nexusreader.NexusReader.NotNexusFileError)
        while not self._nexus_tokenizer.is_eof():
            token = self._nexus_tokenizer.next_token_ucase()
            while token != None and token != 'BEGIN' and not self._nexus_tokenizer.is_eof():
                token = self._nexus_tokenizer.next_token_ucase()
            self._nexus_tokenizer.process_and_clear_comments_for_item(
                    self._global_annotations_target,
                    self.extract_comment_metadata)
            token = self._nexus_tokenizer.next_token_ucase()
            if token == 'TAXA':
                self._parse_taxa_block()
            elif token == 'TREES':
                for tree in self._yield_from_trees_block():
                    yield tree
            elif token == 'BEGIN':
                raise self._nexus_error("'BEGIN' found without completion of previous block",
                        nexusreader.NexusReader.IncompleteBlockError)
            else:
                # unknown block
                token = self._consume_to_end_of_block(token)

    ###########################################################################
    ## Supporting Functions

    def _yield_from_trees_block(self):
        """
        Expectations:
            - current token: "TREES" [part of "BEGIN TREES"]
        """
        token = self._nexus_tokenizer.cast_current_token_to_ucase()
        if token != "TREES":
            raise self._nexus_error("Expecting 'TREES' token, but instead found '{}'".format(token))
        if self.exclude_trees:
            self._consume_to_end_of_block(self._nexus_tokenizer.current_token)
            return
        self._nexus_tokenizer.skip_to_semicolon() # move past "BEGIN TREES" command
        link_title = None
        taxon_namespace = None
        taxon_symbol_mapper = None
        trees_block = None
        block_title = None
        while ((not self._nexus_tokenizer.is_eof())
                and token is not None
                and token != 'END'
                and token != 'ENDBLOCK'):
            token = self._nexus_tokenizer.next_token_ucase()
            if token == 'LINK':
                link_title = self._parse_link_statement().get("taxa")
            elif token == 'TITLE':
                block_title = self._parse_title_statement()
                token = "" # clear; repopulate at start of loop
            elif token == 'TRANSLATE':
                if taxon_namespace is None:
                    taxon_namespace = self._get_taxon_namespace(link_title)
                taxon_symbol_mapper = self._parse_translate_statement(taxon_namespace)
                token = "" # clear; repopulate at start of loop
            elif token == 'TREE':
                if taxon_namespace is None:
                    taxon_namespace = self._get_taxon_namespace(link_title)
                if taxon_symbol_mapper is None:
                    taxon_symbol_mapper = self._get_taxon_symbol_mapper(taxon_namespace=taxon_namespace)
                pre_tree_comments = self._nexus_tokenizer.pull_captured_comments()
                tree_factory = self.tree_factory
                while True:
                    ## After the following, the current token
                    ## will be the token immediately following
                    ## the terminating semi-colon of a tree
                    ## statement. Typically, this will be
                    ## 'TREE' if there is another tree, or
                    ## 'END'/'ENDBLOCK'.
                    tree = self._parse_tree_statement(
                            tree_factory=tree_factory,
                            taxon_symbol_mapper=taxon_symbol_mapper)
                    yield tree
                    if self._nexus_tokenizer.is_eof() or not self._nexus_tokenizer.current_token:
                        break
                    if self._nexus_tokenizer.cast_current_token_to_ucase() != "TREE":
                        token = self._nexus_tokenizer.current_token
                        break
            elif token == 'BEGIN':
                raise self._nexus_error("'BEGIN' found without completion of previous block",
                        nexusreader.NexusReader.IncompleteBlockError)
        self._nexus_tokenizer.skip_to_semicolon() # move past END command
        return

class NexusNewickTreeDataYielder(NexusTreeDataYielder):

    def __init__(self,
            files=None,
            taxon_namespace=None,
            tree_type=None,
            **kwargs):
        kwargs["assume_newick_if_not_nexus"] = kwargs.get("assume_newick_if_not_nexus", True)
        NexusTreeDataYielder.__init__(self,
                files=files,
                taxon_namespace=taxon_namespace,
                tree_type=tree_type,
                **kwargs)



#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Split calculation and management.
DEPRECATED IN DENDROPY 4: USE `dendropy.calculate.treesum` instead.
"""

from dendropy.calculate import treesum
from dendropy.utility import deprecate

class TreeSummarizer(treesum.TreeSummarizer):
    def __init__(self, **kwargs):
        deprecate.dendropy_deprecation_warning(
                preamble="Deprecated since DendroPy 4: The 'dendropy.treesum.TreeSummarizer' class has moved to 'dendropy.calculate.treesum.TreeSummarizer'.",
                old_construct="from dendropy import treesum\nm = treesum.TreeSummarizer(...)",
                new_construct="from dendropy.calculate import treesum\nm = treesum.TreeSummarizer(...)")
        treesum.TreeSummarizer.__init__(self, **kwargs)

class TopologyCounter(treesum.TopologyCounter):
    def __init__(self, **kwargs):
        deprecate.dendropy_deprecation_warning(
                preamble="Deprecated since DendroPy 4: The 'dendropy.treesum.TopologyCounter' class has moved to 'dendropy.calculate.treesum.TopologyCounter'.",
                old_construct="from dendropy import treesum\nm = treesum.TopologyCounter(...)",
                new_construct="from dendropy.calculate import treesum\nm = treesum.TopologyCounter(...)")
        treesum.TopologyCounter.__init__(self, **kwargs)





#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

import os
import sys
import subprocess
from dendropy.utility import error
from dendropy.utility import metavar
from dendropy.utility import libexec
from dendropy.utility import processio
from dendropy.utility import textprocessing

RSCRIPT_EXECUTABLE = os.environ.get(metavar.DENDROPY_RSCRIPT_PATH_ENVAR, "Rscript")
if RSCRIPT_EXECUTABLE == "NONE":
    DENDROPY_RSCRIPT_INTEROPERABILITY = False
else:
    DENDROPY_RSCRIPT_INTEROPERABILITY = True
rsubprocess_pipe_path = libexec.filepath("rsubprocess.R")

class RService(object):

    @staticmethod
    def call(r_commands,
            ignore_error_returncode=False,
            cwd=None,
            env=None,
            rscript_path=RSCRIPT_EXECUTABLE,
            ):
        r"""
        Executes a sequence of commands in R and returns the results. All the
        noise is sunk into the stderr return variable, and just the output
        comes out cleanly in the stdout return variable.

        Parameters
        ----------
        r_commands : iterable of strings
            A list or some other iterable of strings of R commands.
        ignore_error_returncode : bool
            If |True|, then a non-0 return code from the R process will not
            result in an exception being raised.
        cwd : string
            Set the working directory of the R process to this directory.
        env : dictionary
            Environmental variables to set for the R process.
        rscript_path : string
            Path to the Rscript executable.

        Returns
        -------
        returncode : exit value of the R process
        stdout : string
            Contents of the R process standard output.
        stderr : string
            Contents of the R process standard error.

        Examples
        --------

        Build up a script (``s``) to calculate a range of values, print them
        to the standard output, and then post-process this to extract the
        values::

            import itertools
            from dendropy.interop import rstats

            bb = [0.01, 0.05, 0.10, 0.50, 1.0]
            cc = [0.01, 0.05, 0.10, 0.50, 1.0]
            ee = [0.0, 0.1, 0.2]

            # store commands of script as a list
            # to be passed to the ``call()``
            s = []

            # set options, load required libraries, etc.
            s.append("options(digits=22)")
            s.append("library(PBD)")

            # build up list of commands in script
            params = []
            for b, c, e in itertools.product(bb, cc, ee):
                s.append("print(pbd_durspec_mean(pars=c({},{},{})))".format(b, c, e))

            # execute script
            returncode, stdout, stderr  = rstats.call(s)

            # peek at the results
            print(stdout)

            # [1] 69.31472
            # [1] 9.853723
            # [1] 4.981369
            # [1] 0.9950331
            # ...

            # post-process the stdout to extract values
            results = [float(x.split(" ")[1]) for x in stdout.split("\n") if x]

        Notes
        -----

        Note that newlines ('\n') and other special characters will be
        converted before being passed to the R interpreter, so need to
        be escaped or entered as raw string expressions.

        That is, instead of, e.g.::

            returncode, stdout, stderr = RService.call([
                "cat('hello, world\n')",
            ])

        use this::

            returncode, stdout, stderr = RService.call([
                "cat('hello, world\n')",
            ])

        or::

            returncode, stdout, stderr = RService.call([
                r"cat('hello, world\n')",
            ])

        """
        if not textprocessing.is_str_type(r_commands):
            r_commands = "\n".join(r_commands)
        r_commands += "\n"
        invocation_command = [RSCRIPT_EXECUTABLE, rsubprocess_pipe_path]
        p = subprocess.Popen(
                invocation_command,
                stdin=subprocess.PIPE,
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                cwd=cwd,
                env=env,
                )
        stdout, stderr = processio.communicate(p, r_commands)
        if (p.returncode != 0 and not ignore_error_returncode):
            raise error.ExternalServiceError(
                    service_name="Rscript",
                    invocation_command=invocation_command,
                    service_input=r_commands,
                    returncode = p.returncode,
                    stdout=stdout,
                    stderr=stderr)
        return p.returncode, stdout, stderr

def call(*args, **kwargs):
    return RService.call(*args, **kwargs)




#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Models and modeling of discrete character evolution.
"""

import copy
import math
import itertools
from dendropy.utility import GLOBAL_RNG
from dendropy.calculate import probability
import dendropy

############################################################################
## Character Evolution Modeling

class DiscreteCharacterEvolutionModel(object):
    "Base class for discrete character substitution models."

    def __init__(self, state_alphabet, stationary_freqs=None, rng=None):
        """
        __init__ initializes the state_alphabet to define the character type on which
        this model acts.  The objects random number generator will be ``rng`` or 'GLOBAL_RNG'
        """
        self.state_alphabet = state_alphabet
        if rng is None:
            self.rng = GLOBAL_RNG
        else:
            self.rng = rng

    def pmatrix(self, tlen, rate=1.0):
        """
        Returns a matrix of nucleotide substitution
        probabilities.
        """
        raise NotImplementedError

    def simulate_descendant_states(self,
        ancestral_states,
        edge_length,
        mutation_rate=1.0,
        rng=None):
        """
        Returns descendent sequence given ancestral sequence.
        """
        if rng is None:
            rng = self.rng
        pmat = self.pmatrix(edge_length, mutation_rate)
        multi = probability.sample_multinomial
        desc_states = []
        for state in ancestral_states:
            anc_state_idx = state.index
            desc_state_idx = multi(pmat[anc_state_idx], rng)
            desc_states.append(self.state_alphabet[desc_state_idx])
        return desc_states

class DiscreteCharacterEvolver(object):
    "Evolves sequences on a tree."

    def __init__(self,
     seq_model=None,
     mutation_rate=None,
     seq_attr='sequences',
     seq_model_attr="seq_model",
     edge_length_attr="length",
     edge_rate_attr="mutation_rate",
     seq_label_attr='taxon'):
        "__init__ sets up meta-data dealing with object nomenclature and semantics."
        self.seq_model = seq_model
        self.mutation_rate = mutation_rate
        self.seq_attr = seq_attr
        self.seq_model_attr = seq_model_attr
        self.edge_length_attr = edge_length_attr
        self.edge_rate_attr = edge_rate_attr
        self.seq_label_attr = seq_label_attr

    def evolve_states(self,
            tree,
            seq_len,
            root_states=None,
            simulate_root_states=True,
            in_place=True,
            rng=None):
        """
        Appends a new sequence of length ``seq_len`` to a list at each node
        in ``tree``.  The attribute name of this list in each node is given
        by ``seq_attr``. If ``seq_model`` is None, ``tree.seq_model`` or
        ``seq_model`` at each node must be specified. If ``in_place`` is
        False, the tree is copied first, otherwise original tree is modified.
        If ``root_states`` is given, this will be used as the sequence for the root.
        If not, and if ``simulate_root_states`` is True, then the sequence for the
        root will be drawn from the stationary distribution of the character model.
        """
        if rng is None:
            rng = GLOBAL_RNG
        if not in_place:
            tree = tree.clone(1) # ==> taxon_namespace_scoped_copy()

        if self.seq_model is None:
            seq_model = getattr(tree, self.seq_model_attr, None)

        # loop through edges in preorder (root->tips)
        n_prev_seq = None  # to mollify linter undefined variable warning
        for edge in tree.preorder_edge_iter():
            node = edge.head_node
            if not hasattr(node, self.seq_attr):
                setattr(node, self.seq_attr, [])
            seq_list = getattr(node, self.seq_attr)
            if edge.tail_node:
                par = edge.tail_node
                assert n_prev_seq is not None
                if len(seq_list) != n_prev_seq:
                    raise ValueError("'%s' length varies among nodes" % self.seq_attr)
                par_seq = getattr(par, self.seq_attr)[-1]
                seq_model  = getattr(edge, self.seq_model_attr, None) or self.seq_model
                length = getattr(edge, self.edge_length_attr)
                mutation_rate = getattr(edge, self.edge_rate_attr, None) or self.mutation_rate
                seq_list.append(seq_model.simulate_descendant_states(par_seq,
                                                                     length,
                                                                     mutation_rate,
                                                                     rng=rng))
            else:
                # no tail node: root
                n_prev_seq = len(seq_list)
                if root_states is not None:
                    seq_list.append(root_states)
                elif simulate_root_states:
                    seq_model  = getattr(node.edge, self.seq_model_attr, None) or self.seq_model
                    seq_list.append(seq_model.stationary_sample(seq_len, rng=rng))
                else:
                    assert n_prev_seq > 0
                    n_prev_seq -= 1
        return tree

    def extend_char_matrix_with_characters_on_tree(self,
            char_matrix,
            tree,
            include=None,
            exclude=None):
        """
        Creates a character matrix with new sequences (or extends sequences of
        an existing character matrix if provided via ``char_matrix``),
        where the the sequence for each taxon corresponds to the concatenation
        of all sequences in the list of sequences associated with tip that
        references the given taxon.
        Specific sequences to be included/excluded can be fine-tuned using the
        ``include`` and ``exclude`` args, where ``include=None`` means to include all
        by default, and ``exclude=None`` means to exclude all by default.
        """
        for leaf in tree.leaf_nodes():
            cvec = char_matrix[leaf.taxon]
            seq_list = getattr(leaf, self.seq_attr)
            for seq_idx, seq in enumerate(seq_list):
                if ((include is None) or (seq_idx in include))  \
                    and ((exclude is None) or (seq_idx not in exclude)):
                    for state in seq:
                        cvec.append(state)
        return char_matrix

    def clean_tree(self, tree):
        for nd in tree:
            # setattr(nd, self.seq_attr, [])
            delattr(nd, self.seq_attr)

############################################################################
## Specialized Models: nucldeotides

class NucleotideCharacterEvolutionModel(DiscreteCharacterEvolutionModel):
    "General nucleotide substitution model."

    def __init__(self, base_freqs=None, state_alphabet=None, rng=None):
        "__init__ calls SeqModel.__init__ and sets the base_freqs field"
        if state_alphabet is None:
            state_alphabet = dendropy.DNA_STATE_ALPHABET
        DiscreteCharacterEvolutionModel.__init__(
                self,
                state_alphabet=state_alphabet,
                rng=rng)
        if base_freqs is None:
            self.base_freqs = [0.25, 0.25, 0.25, 0.25]
        else:
            self.base_freqs = base_freqs

    def stationary_sample(self, seq_len, rng=None):
        """
        Returns a NucleotideSequence() object with length ``length``
        representing a sample of characters drawn from this model's
        stationary distribution.
        """
        probs = self.base_freqs
        char_state_indices = [probability.sample_multinomial(probs, rng) for i in range(seq_len)]
        return [self.state_alphabet[idx] for idx in char_state_indices]

    def is_purine(self, state_index):
        """
        Returns True if state_index represents a purine (A or G) row or column
        index: 0, 2
        """
        return state_index % 2 == 0

    def is_pyrimidine(self, state_index):
        """
        Returns True if state_index represents a pyrimidine (C or T) row or column
        index: 1, 3
        """
        return state_index % 2 == 1

    def is_transversion(self, state1_idx, state2_idx):
        """
        Returns True if the change from state1 to state2, as
        represented by the row or column indices, is a transversional
        change.
        """
        return (self.is_purine(state1_idx) and self.is_pyrimidine(state2_idx)) \
               or (self.is_pyrimidine(state1_idx) and self.is_purine(state2_idx))

    def is_purine_transition(self, state1_idx, state2_idx):
        """
        Returns True if the change from state1 to state2, as
        represented by the row or column indices, is a purine
        transitional change.
        """
        return self.is_purine(state1_idx) and self.is_purine(state2_idx)

    def is_pyrimidine_transition(self, state1_idx, state2_idx):
        """
        Returns True if the change from state1 to state2, as
        represented by the row or column indices, is a pyrimidine
        transitional change.
        """
        return self.is_pyrimidine(state1_idx) \
               and self.is_pyrimidine(state2_idx)

    def is_transition(self, state1_idx, state2_idx):
        """
        Returns True if the change from state1 to state2, as
        represented by the row or column indices, is a
        transitional change.
        """
        return (self.is_purine(state1_idx) and self.is_purine(state2_idx)) \
               or (self.is_pyrimidine(state1_idx) and self.is_pyrimidine(state2_idx))

class Hky85(NucleotideCharacterEvolutionModel):
    """
    Hasegawa et al. 1985 model. Implementation following Swofford et
    al., 1996.
    """

    def __init__(self, kappa=1.0, base_freqs=None, state_alphabet=None, rng=None):
        "__init__: if no arguments given, defaults to JC69."
        if state_alphabet is None:
            state_alphabet = dendropy.DNA_STATE_ALPHABET
        NucleotideCharacterEvolutionModel.__init__(
                self,
                base_freqs=base_freqs,
                state_alphabet=state_alphabet,
                rng=rng)
        self.correct_rate = True
        self.kappa = kappa

    def __repr__(self):
        rep = "kappa=%f bases=%s" % (self.kappa, str(self.base_freqs))
        return rep

    def corrected_substitution_rate(self, rate):
        """Returns the factor that we have to multiply to the branch length
        to make branch lengths proportional to # of substitutions per site."""
        if self.correct_rate:
            pia = self.base_freqs[0]
            pic = self.base_freqs[1]
            pig = self.base_freqs[2]
            pit = self.base_freqs[3]
            f = self.kappa*(pia*pig + pic*pit)
            f += (pia + pig)*(pic + pit)
            return (rate * 0.5/f)  # (rate * 0.5/f)
        else:
            return rate

    def pij(self, state_i, state_j, tlen, rate=1.0):
        """
        Returns probability, p_ij, of going from state i to state j
        over time tlen at given rate. (tlen * rate = nu, expected
        number of substitutions)
        """
        nu = self.corrected_substitution_rate(rate) * tlen
        if self.is_purine(state_j):
            sumfreqs = self.base_freqs[0] + self.base_freqs[2]
        else:
            sumfreqs = self.base_freqs[1] + self.base_freqs[3]
        factorA = 1 + (sumfreqs * (self.kappa - 1.0))
        if state_i == state_j:
            pij = self.base_freqs[state_j] \
                  + self.base_freqs[state_j] \
                      * (1.0/sumfreqs - 1) * math.exp(-1.0 * nu) \
                  + ((sumfreqs - self.base_freqs[state_j])/sumfreqs) \
                      * math.exp(-1.0 * nu * factorA)

        elif self.is_transition(state_i, state_j):
            pij = self.base_freqs[state_j] \
                  + self.base_freqs[state_j] \
                      * (1.0/sumfreqs - 1) * math.exp(-1.0 * nu) \
                  - (self.base_freqs[state_j] / sumfreqs) \
                      * math.exp(-1.0 * nu * factorA)
        else:
            pij = self.base_freqs[state_j] * (1.0 - math.exp(-1.0 * nu))
        return pij

    def qmatrix(self, rate=1.0):
        "Returns the instantaneous rate of change matrix."
        rate = self.corrected_substitution_rate(rate)
        qmatrix = []
        for state_i in range(4):
            qmatrix.append([])
            for state_j in range(4):
                if state_i == state_j:
                    # we cheat here and insert a placeholder till the
                    # other cells are calculated
                    qij = 0.0
                else:
                    if self.is_transition(state_i, state_j):
                        qij = rate * self.kappa * self.base_freqs[state_j]
                    else:
                        qij = rate * self.base_freqs[state_j]
                qmatrix[state_i].append(qij)
        for state in range(4):
            qmatrix[state][state] = -1.0 * sum(qmatrix[state])
        return qmatrix

    def pvector(self, state, tlen, rate=1.0):
        """
        Returns a vector of transition probabilities for a given state
        over time ``tlen`` at rate ``rate`` for ``state``. (tlen * rate =
        nu, expected number of substitutions)
        """
        pvec = []
        # in case later we want to allow characters passed in here
        state_i = state
        for state_j in range(4):
            pvec.append(self.pij(state_i, state_j, tlen=tlen, rate=rate))
        return pvec

    def pmatrix(self, tlen, rate=1.0):
        """
        Returns a matrix of nucleotide substitution
        probabilities. Based on analytical solution by Swofford et
        al., 1996. (tlen * rate = nu, expected number of
        substitutions)
        """
        pmatrix = []
        for state_i in range(4):
            pmatrix.append(self.pvector(state_i, tlen=tlen, rate=rate))
        return pmatrix

class Jc69(Hky85):
    """
    Jukes-Cantor 1969 model. Specializes HKY85 such that
    kappa = 1.0, and base frequencies = [0.25, 0.25, 0.25, 0.25].
    """
    def __init__(self, state_alphabet=None, rng=None):
        "__init__: uses Hky85.__init__"
        Hky85.__init__(self,
                kappa=1.0,
                base_freqs=[0.25, 0.25, 0.25, 0.25],
                state_alphabet=state_alphabet,
                rng=rng,
                )



##############################################################################
## Wrappers for Convenience

def simulate_discrete_char_dataset(seq_len,
        tree_model,
        seq_model,
        mutation_rate=1.0,
        root_states=None,
        dataset=None,
        rng=None):
    """
    Wrapper to conveniently generate a DataSet simulated under
    the given tree and character model.

    Parameters
    ----------

    seq_len       : int
        Length of sequence (number of characters).
    tree_model    : |Tree|
        Tree on which to simulate.
    seq_model     : dendropy.model.discrete.DiscreteCharacterEvolutionModel
        The character substitution model under which to to evolve the
        characters.
    mutation_rate : float
        Mutation *modifier* rate (should be 1.0 if branch lengths on tree
        reflect true expected number of changes).
    root_states``   : list
        Vector of root states (length must equal ``seq_len``).
    dataset       : |DataSet|
        If given, the new dendropy.CharacterMatrix object will be
        added to this (along with a new taxon_namespace if
        required). Otherwise, a new dendropy.DataSet
        object will be created.
    rng           : random number generator
        If not given, 'GLOBAL_RNG' will be used.

    Returns
    -------

    d : |DataSet|

    """
    if dataset is None:
        dataset = dendropy.DataSet()
    if tree_model.taxon_namespace not in dataset.taxon_namespaces:
        taxon_namespace = dataset.add_taxon_namespace(tree_model.taxon_namespace)
    else:
        taxon_namespace = tree_model.taxon_namespace
    char_matrix = simulate_discrete_chars(
        seq_len=seq_len,
        tree_model=tree_model,
        seq_model=seq_model,
        mutation_rate=mutation_rate,
        root_states=root_states,
        char_matrix=None,
        rng=None)
    dataset.add_char_matrix(char_matrix=char_matrix)
    return dataset

def simulate_discrete_chars(
        seq_len,
        tree_model,
        seq_model,
        mutation_rate=1.0,
        root_states=None,
        char_matrix=None,
        retain_sequences_on_tree=False,
        rng=None):
    """
    Wrapper to conveniently generate a characters simulated under
    the given tree and character model.

    Since characters will be appended to existing sequences, you can simulate a
    sequences under a mixed model by calling this method multiple times with
    different character models and/or different mutation rates, passing
    in the same ``char_matrix`` object each time.

    Parameters
    ----------

    seq_len       : int
        Length of sequence (number of characters).
    tree_model    : |Tree|
        Tree on which to simulate.
    seq_model     : dendropy.model.discrete.DiscreteCharacterEvolutionModel
        The character substitution model under which to to evolve the
        characters.
    mutation_rate : float
        Mutation *modifier* rate (should be 1.0 if branch lengths on tree
        reflect true expected number of changes).
    root_states``   : list
        Vector of root states (length must equal ``seq_len``).
    char_matrix   : |DnaCharacterMatrix|
        If given, new sequences for taxa on ``tree_model`` leaf_nodes will be
        appended to existing sequences of corresponding taxa in char_matrix; if
        not, a new |DnaCharacterMatrix| object will be created.
    retain_sequences_on_tree : bool
        If |False|, sequence annotations will be cleared from tree after
        simulation. Set to |True| if you want to, e.g., evolve and accumulate
        different sequences on tree, or retain information for other purposes.
    rng           : random number generator
        If not given, 'GLOBAL_RNG' will be used.

    Returns
    -------
    d : a dendropy.datamodel.CharacterMatrix object.

    """
    seq_evolver = DiscreteCharacterEvolver(seq_model=seq_model,
                               mutation_rate=mutation_rate)
    tree = seq_evolver.evolve_states(
        tree=tree_model,
        seq_len=seq_len,
        root_states=None,
        rng=rng)
    if char_matrix is None:
        char_matrix = dendropy.DnaCharacterMatrix(taxon_namespace=tree_model.taxon_namespace)
        char_matrix.taxon_namespace = tree_model.taxon_namespace
    else:
        assert char_matrix.taxon_namespace is tree_model.taxon_namespace, "conflicting taxon sets"
    seq_evolver.extend_char_matrix_with_characters_on_tree(
            char_matrix=char_matrix,
            tree=tree)
    if not retain_sequences_on_tree:
        seq_evolver.clean_tree(tree)
    return char_matrix

def hky85_chars(
        seq_len,
        tree_model,
        mutation_rate=1.0,
        kappa=1.0,
        base_freqs=[0.25, 0.25, 0.25, 0.25],
        root_states=None,
        char_matrix=None,
        retain_sequences_on_tree=False,
        rng=None):
    """
    Convenience class to wrap generation of characters (as a CharacterBlock
    object) based on the HKY model.

    Parameters
    ----------

    seq_len       : int
        Length of sequence (number of characters).
    tree_model    : |Tree|
        Tree on which to simulate.
    mutation_rate : float
        Mutation *modifier* rate (should be 1.0 if branch lengths on tree
        reflect true expected number of changes).
    root_states``   : list
        Vector of root states (length must equal ``seq_len``).
    char_matrix   : |DnaCharacterMatrix|
        If given, new sequences for taxa on ``tree_model`` leaf_nodes will be
        appended to existing sequences of corresponding taxa in char_matrix; if
        not, a new |DnaCharacterMatrix| object will be created.
    retain_sequences_on_tree : bool
        If |False|, sequence annotations will be cleared from tree after
        simulation. Set to |True| if you want to, e.g., evolve and accumulate
        different sequences on tree, or retain information for other purposes.
    rng           : random number generator
        If not given, 'GLOBAL_RNG' will be used.

    Returns
    -------
    d : |DnaCharacterMatrix|
        The simulated alignment.

    Since characters will be appended to existing sequences, you can simulate a
    sequences under a mixed model by calling this method multiple times with
    different character model parameter values and/or different mutation
    rates, passing in the same ``char_matrix`` object each time.
    """
    if char_matrix is None:
        char_matrix = dendropy.DnaCharacterMatrix(taxon_namespace=tree_model.taxon_namespace)
    else:
        assert char_matrix.taxon_namespace is tree_model.taxon_namespace
    state_alphabet = char_matrix.default_state_alphabet
    seq_model = Hky85(
            kappa=kappa,
            base_freqs=base_freqs,
            state_alphabet=state_alphabet,
            rng=rng,
    )
    return simulate_discrete_chars(seq_len=seq_len,
                               tree_model=tree_model,
                               seq_model=seq_model,
                               mutation_rate=mutation_rate,
                               root_states=root_states,
                               char_matrix=char_matrix,
                               rng=rng)




#! /usr/bin/env python
# -*- coding: utf-8 -*-

##############################################################################
##  DendroPy Phylogenetic Computing Library.
##
##  Copyright 2010-2015 Jeet Sukumaran and Mark T. Holder.
##  All rights reserved.
##
##  See "LICENSE.rst" for terms and conditions of usage.
##
##  If you use this work or any portion thereof in published work,
##  please cite it as:
##
##     Sukumaran, J. and M. T. Holder. 2010. DendroPy: a Python library
##     for phylogenetic computing. Bioinformatics 26: 1569-1571.
##
##############################################################################

"""
Models, modeling and model-fitting of parsimony.
"""

from functools import reduce
import operator
import dendropy
from dendropy.utility.error import TaxonNamespaceIdentityError

class _NodeStateSetMap(dict):
    def __init__(self, taxon_state_sets_map=None):
        self.taxon_state_sets_map = taxon_state_sets_map
    def __getitem__(self, key):
        try:
            return dict.__getitem__(self, key)
        except KeyError:
            v = self.taxon_state_sets_map[key.taxon]
            self[key] = v
            return v

def _store_sets_as_attr(n, state_sets_attr_name, v):
    setattr(n, state_sets_attr_name, v)

def _retrieve_state_sets_from_attr(n, state_sets_attr_name, taxon_state_sets_map):
    try:
        return getattr(n, state_sets_attr_name)
    except AttributeError:
        v = taxon_state_sets_map[n.taxon]
        setattr(n, state_sets_attr_name, v)
        return v

def fitch_down_pass(
        postorder_nodes,
        state_sets_attr_name="state_sets",
        taxon_state_sets_map=None,
        weights=None,
        score_by_character_list=None,
        ):
    """
    Returns the parsimony score given a list of nodes in postorder and
    associated states, using Fitch's (1971) unordered parsimony algorithm.

    Parameters
    ----------
    postorder_nodes : iterable of/over |Node| objects
        An iterable of |Node| objects in in order of post-order
        traversal of the tree.
    state_sets_attr_name : str
        Name of attribute on |Node| objects in which state set lists
        will stored/accessed. If |None|, then state sets will not be stored on
        the tree.
    taxon_state_sets_map : dict[taxon] = state sets
        A dictionary that takes a taxon object as a key and returns a state set
        list as a value. This will be used to populate the state set of a node
        that has not yet had its state sets scored and recorded (typically,
        leaves of a tree that has not yet been processed).
    weights : iterable
        A list of weights for each pattern.
    score_by_character_list : None or list
        If not |None|, should be a reference to a list object.
        This list will be populated by the scores on a character-by-character
        basis.

    Returns
    -------
    s : int
        Parismony score of tree.

    Notes
    -----
    Currently this requires a bifurcating tree (even at the root).

    Examples
    --------

    Assume that we have a tree, ``tree``, and an associated data set, ``data``::

        import dendropy
        from dendropy.model.parsimony import fitch_down_pass

        taxa = dendropy.TaxonNamespace()
        data = dendropy.StandardCharacterMatrix.get_from_path(
                "apternodus.chars.nexus",
                "nexus",
                taxon_namespace=taxa)
        tree = dendropy.Tree.get_from_path(
                "apternodus.tre",
                "nexus",
                taxon_namespace=taxa)
        taxon_state_sets_map = data.taxon_state_sets_map(gaps_as_missing=True)

    The following will return the parsimony score of the ``tree`` with
    respect to the data in ``data``::

        score = fitch_down_pass(
                nodes=tree.postorder_node_iter(),
                taxon_state_sets_map=taxon_set_map)
        print(score)

    In the above, every |Node| object of ``tree`` will have an attribute
    added, "state_sets", that stores the list of state sets from the analysis::

        for nd in tree:
            print(nd.state_sets)

    If you want to store the list of state sets in a different attribute, e.g.,
    "analysis1_states"::

        score = fitch_down_pass(
                nodes=tree.postorder_node_iter(),
                state_sets_attr_name="analysis1_states",
                taxon_state_sets_map=taxon_set_map)
        print(score)
        for nd in tree:
            print(nd.analysis1_states)

    Or not to store these at all::

        score = fitch_down_pass(
                nodes=tree.postorder_node_iter(),
                state_sets_attr_name=None,
                taxon_state_sets_map=taxon_set_map)
        print(score)

    Scoring custom data can be done by something like the following::

        taxa = dendropy.TaxonNamespace()
        taxon_state_sets_map = {}
        t1 = taxa.require_taxon("A")
        t2 = taxa.require_taxon("B")
        t3 = taxa.require_taxon("C")
        t4 = taxa.require_taxon("D")
        t5 = taxa.require_taxon("E")
        taxon_state_sets_map[t1] = [ set([0,1]),  set([0,1]),  set([0]),     set([0]) ]
        taxon_state_sets_map[t2] = [ set([1]),    set([1]),    set([1]),     set([0]) ]
        taxon_state_sets_map[t3] = [ set([0]),    set([1]),    set([1]),     set([0]) ]
        taxon_state_sets_map[t4] = [ set([0]),    set([1]),    set([0,1]),   set([1]) ]
        taxon_state_sets_map[t5] = [ set([1]),    set([0]),    set([1]),     set([1]) ]
        tree = dendropy.Tree.get_from_string(
                "(A,(B,(C,(D,E))));", "newick",
                taxon_namespace=taxa)
        score = fitch_down_pass(tree.postorder_node_iter(),
                taxon_state_sets_map=taxon_state_sets_map)
        print(score)

    """
    if score_by_character_list is not None:
        assert len(score_by_character_list) == 0
        for idx in range(len(list(taxon_state_sets_map.values())[0])): # this is unacceptable!
            score_by_character_list.append(0)
    score = 0
    if state_sets_attr_name is None:
        node_state_set_map = _NodeStateSetMap(taxon_state_sets_map)
        get_node_state_sets = lambda node : node_state_set_map[node]
        set_node_state_sets = lambda node, v : node_state_set_map.__setitem__(node, v)
    else:
        get_node_state_sets = lambda node : _retrieve_state_sets_from_attr(node, state_sets_attr_name, taxon_state_sets_map)
        set_node_state_sets = lambda node, v : _store_sets_as_attr(node, state_sets_attr_name, v)
    for nd in postorder_nodes:
        c = nd.child_nodes()
        if not c:
            ss = get_node_state_sets(nd)
            continue
        left_c, right_c = c[:2]
        remaining = c[2:]
        left_ssl = get_node_state_sets(left_c)
        while True:
            right_ssl = get_node_state_sets(right_c)
            result = []
            for n, ssp in enumerate(zip(left_ssl, right_ssl)):
                left_ss, right_ss = ssp
                inter = left_ss.intersection(right_ss)
                if inter:
                    result.append(inter)
                else:
                    if weights is None:
                        wt = 1
                    else:
                        wt = weights[n]
                    score += wt
                    result.append(left_ss.union(left_ss, right_ss))
                    if score_by_character_list is not None:
                        # try:
                        #     score_by_character_list[n] += wt
                        # except IndexError:
                        #     score_by_character_list.append(wt)
                        score_by_character_list[n] += wt
            if remaining:
                right_c = remaining.pop(0)
                left_ssl = result
            else:
                break
        # setattr(nd, state_sets_attr_name, result)
        set_node_state_sets(nd, result)
    return score

def fitch_up_pass(
        preorder_node_list,
        state_sets_attr_name="state_sets",
        taxon_state_sets_map=None):
    """
    Finalizes the state set lists associated with each node using the "final
    phase" of Fitch's (1971) unordered parsimony algorithm.

    Parameters
    ----------
    postorder_nodes : iterable of/over |Node| objects
        An iterable of |Node| objects in in order of post-order
        traversal of the tree.
    state_sets_attr_name : str
        Name of attribute on |Node| objects in which state set lists
        will stored/accessed. If |None|, then state sets will not be stored on
        the tree.
    taxon_state_sets_map : dict[taxon] = state sets
        A dictionary that takes a taxon object as a key and returns a state set
        list as a value. This will be used to populate the state set of a node
        that has not yet had its state sets scored and recorded (typically,
        leaves of a tree that has not yet been processed).

    Notes
    -----
    Currently this requires a bifurcating tree (even at the root).

    Examples
    --------

    ::

        taxa = dendropy.TaxonNamespace()
        data = dendropy.StandardCharacterMatrix.get_from_path(
                "apternodus.chars.nexus",
                "nexus",
                taxon_namespace=taxa)
        tree = dendropy.Tree.get_from_path(
                "apternodus.tre",
                "nexus",
                taxon_namespace=taxa)
        taxon_state_sets_map = data.taxon_state_sets_map(gaps_as_missing=True)
        score = fitch_down_pass(tree.postorder_node_iter(),
                taxon_state_sets_map=taxon_state_sets_map)
        print(score)
        fitch_up_pass(tree.preorder_node_iter())
        for nd in tree:
            print(nd.state_sets)

    """
    node_state_sets_map = {}
    for nd in preorder_node_list:
        c = nd.child_nodes()
        p = nd.parent_node
        if (not c) or (not p):
            continue
        assert(len(c) == 2)
        left_c, right_c = c
        try:
            left_ssl = getattr(left_c, state_sets_attr_name)
        except AttributeError:
            if not taxon_state_sets_map:
                raise
            left_ssl = taxon_state_sets_map[left_c.taxon]
        try:
            right_ssl = getattr(right_c, state_sets_attr_name)
        except AttributeError:
            if not taxon_state_sets_map:
                raise
            right_ssl = taxon_state_sets_map[right_c.taxon]
        par_ssl = getattr(p, state_sets_attr_name)
        curr_ssl = getattr(nd, state_sets_attr_name)
        result = []
        for n, ssp in enumerate(zip(par_ssl, curr_ssl, left_ssl, right_ssl)):
            par_ss, curr_ss, left_ss, right_ss = ssp

            down_parup_inter = par_ss.intersection(curr_ss)
            if down_parup_inter == par_ss:
                final_ss = down_parup_inter
            else:
                rl_inter = left_ss.intersection(right_ss)
                if not rl_inter:
                    final_ss = par_ss.union(curr_ss)
                else:
                    in_par_and_left = par_ss.intersection(left_ss)
                    in_par_and_right = par_ss.intersection(right_ss)
                    final_ss = in_par_and_left.union(in_par_and_right, curr_ss)
            #_LOG.debug("downpass = %s, par = %s, left = %s, right = %s, final_ss= %s" %
            #                    (str(curr_ss), str(par_ss), str(left_ss), str(right_ss), str(final_ss)))
            result.append(final_ss)
        setattr(nd, state_sets_attr_name, result)


def parsimony_score(
        tree,
        chars,
        gaps_as_missing=True,
        weights=None,
        score_by_character_list=None,
        ):
    """
    Calculates the score of a tree, ``tree``, given some character data,
    ``chars``, under the parsimony model using the Fitch algorithm.

    Parameters
    ----------
    tree : a |Tree| instance
        A |Tree| to be scored. Must reference the same |TaxonNamespace| as
        ``chars``.
    chars : a |CharacterMatrix| instance
        A |CharacterMatrix|-derived object with data to be scored. Must have
        the same |TaxonNamespace| as ``tree``.
    gap_as_missing : bool
        If |True| [default], then gaps will be treated as missing data.
        If |False|, then gaps will be treated as a new/additional state.
    weights : iterable
        A list of weights for each pattern/column in the matrix.
    score_by_character_list : None or list
        If not |None|, should be a reference to a list object.
        This list will be populated by the scores on a character-by-character
        basis.

    Returns
    -------
    pscore : int
        The parsimony score of the tree given the data.

    Examples
    --------

    ::

        import dendropy
        from dendropy.calculate import treescore

        # establish common taxon namespace
        taxon_namespace = dendropy.TaxonNamespace()

        # Read data; if data is, e.g., "standard", use StandardCharacterMatrix.
        # If unsure of data type, can do:
        #       dataset = dendropy.DataSet.get(
        #               path="path/to/file.nex",
        #               schema="nexus",
        #               taxon_namespace=tns,)
        #       chars = dataset.char_matrices[0]
        chars = dendropy.DnaCharacterMatrix.get(
                path="pythonidae.chars.nexus",
                schema="nexus",
                taxon_namespace=taxon_namespace)
        tree = dendropy.Tree.get(
                path="pythonidae.mle.newick",
                schema="newick",
                taxon_namespace=taxon_namespace)

        # We store the site-specific scores here
        # This is optional; if we do not want to
        # use the per-site scores, just pass in |None|
        # for the ``score_by_character_list`` argument
        # or do not specify this argument at all.
        score_by_character_list = []

        score = treescore.parsimony_score(
                tree,
                chars,
                gaps_as_missing=False,
                score_by_character_list=score_by_character_list)

        # Print the results: the score
        print("Score: {}".format(score))

        # Print the results: the per-site scores
        for idx, x in enumerate(score_by_character_list):
            print("{}: {}".format(idx+1, x))

    Notes
    -----

    If the same data is going to be used to score multiple trees or multiple times,
    it is probably better to generate the 'taxon_state_sets_map' once and call
    "fitch_down_pass" directly yourself, as this function generates a new map
    each time.

    """
    if tree.taxon_namespace is not chars.taxon_namespace:
        raise TaxonNamespaceIdentityError(tree, chars)
    taxon_state_sets_map = chars.taxon_state_sets_map(gaps_as_missing=gaps_as_missing)
    nodes = tree.postorder_node_iter()
    pscore = fitch_down_pass(nodes,
            taxon_state_sets_map=taxon_state_sets_map,
            weights=weights,
            score_by_character_list=score_by_character_list)
    return pscore

